/*
 * Dorrit Copyright (c) 2009, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Dorrit is a minimal, multiplexing, kqueue-based, partially-compliant
 * HTTP/1.1 server, which can only serve static content.  It is a variant
 * of the drood web server, specialized to be faster at serving static
 * content.  The only parts of HTTP/1.1 it implements are persistent
 * connections, the "gzip" content-encoding, and name-based virtual hosting
 * dictated by the value of the "Host" request header.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <dirent.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>
#include <regex.h>
#include <string.h>
#include <ctype.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.3"
#define PIDFILE "/var/run/dorrit.pid"
#define TYPEFILE DATADIR "/types.tab"

/*
 * Size of the output queue used by kevent().
 */

#define QLEN 1024

/*
 * This is size of the IO buffers used in connection control blocks.
 * Modifying this value may require a corresponding modification to
 * the precision of the chunk header conversion performed in
 * respond_chunked().  32K is the default size of a socket send buffer
 * on i386 FreeBSD 6.x.  We try to move data in chunks of this size in an
 * attempt to "fill the pipe" of each connection.
 *
 * This should never be set to less than 2000.
 */

#define IOBUFSIZE 32768

/*
 * This is the size of the input buffers used in ccbs.
 */

#define INBUFSIZE 512

/*
 * A connection is in one of the following states at any one time.  There
 * are handlers for each state which are invoked when kevent() returns an
 * event on a connection.  A connection may have the handler for its
 * current state called more than once as part of the multiplexing process.
 * Eventually a handler will advance a connection to a succeeding state,
 * unless a system error occurs, or the connection times-out, when it will
 * be dropped.  Any connection in a state described as a terminal state,
 * will transition back to the RECEIVE state for a persistent connection,
 * otherwise the connection will be closed.

 * RECEIVE:  Receive request header.  Transitions to HEADER or ERROR.

 * HEADER:   Send response header.  Is terminal state for HEAD requests.
 *           When a response body follows the header, transitions to
 *           RESPOND.

 * RESPOND:  Send response body.
 *           Invoked at least once for each IOBUFSIZE portion of requested
 *           resource.  Terminal state.

 * ERROR:    Send an error response header and body.  Terminal state.

 */

/*
 * connection control block.
 */

struct ccb
{
   int type, verno, error, close, count, sock, source, inactive, timer, gzip;

   int request_found, close_found, gzip_found,
      host_found, since_found, unmodified_found;

   void ( *state )( struct ccb * );

   off_t size, written;

   char *request, *method, *agent, *host, *version, *addr,
      *since, *unmodified;

   char buffer[ IOBUFSIZE ], inbuf[ INBUFSIZE ], *inptr;
};

void receive_request( struct ccb * );
void analyze_request( struct ccb * );
void respond_error( struct ccb * );
void respond_header( struct ccb * );
void respond_static( struct ccb * );

void ( *RECEIVE  )( struct ccb * ) = receive_request;
void ( *RESPOND  )( struct ccb * ) = respond_static;
void ( *ERROR    )( struct ccb * ) = respond_error;
void ( *HEADER   )( struct ccb * ) = respond_header;

/*
 * These must start at 1.  0 means "no error".
 */

#define ERR_BADVERSION  1
#define ERR_BADREQUEST  2
#define ERR_FORBIDDEN   3
#define ERR_BADMETHOD   4
#define ERR_NOTFOUND    5
#define ERR_INTERNAL    6
#define ERR_NOCONTENT   7
#define ERR_NOTMODIFIED 8
#define ERR_MODIFIED    9

/*
 * KQueue Input and output queues.  We can add a maximun of four events to
 * the input queue for every item in the output queue, while processing
 * the output queue, except when the output event is for the listening
 * socket.  We can accept up to 100 connections when the listening socket
 * has them ready to go.  Each added connection adds three events to the
 * input queue.
 */

#define INQ QLEN * 4 + 300
struct kevent inqueue[ INQ ];
struct kevent outqueue[ QLEN ];

int in = 0, out = 0, closed = 0, read_timeout = 20, write_timeout = 60,
   instances = 1;

extern char *optarg;
char *root_resource = "index.html",
     *server_header = "Server: Dorrit/" VERSION "\r\n";

int fd = -1, increment = 1024,
   matched = 0, logging = 0, testing = 0, do_chroot = 0,
   backlog = 1000;

struct passwd *passwd;
struct group *group;

char *interface = "", *port = "80", *grp = "nobody", *user = "nobody",
   *root = NULL, *default_type = "text/plain; charset=utf-8";

/*
 * Stack which grows dynamically.
 */

union stack_u {
   void *ptr;
   pid_t pid;
};

struct stack
{
   int free, used;
   union stack_u *top, *values;
} *type_table = NULL, *children = NULL;

char *errstrings[ 11 ] = { "",
     " 505 HTTP Version Not Supported\r\n",
     " 400 Bad Request\r\n",
     " 403 Forbidden\r\n",
     " 405 Method Not Allowed\r\nAllow: GET, HEAD\r\n",
     " 404 Not Found\r\n",
     " 500 Internal Server Error\r\n",
     " 204 No Content\r\n",
     " 304 Not Modified\r\n",
     " 402 Precondition Failed\r\n"
};

regex_t request_rx, host_rx, gzip_rx,
   close_rx, parent_rx, suffix_rx, type_rx, since_rx, unmodified_rx;


/*
 * String which grows dynamically.
 */

struct string
{
   int free, used;
   char *top, *str;
};

void ev_set( int desc, short filter, u_short flags, struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = desc;
   kev->filter = filter;
   kev->fflags = 0;
   kev->flags = flags;
   kev->udata = item;
}

void flip_events( struct ccb *item )
{
   ev_set( item->sock, EVFILT_READ, EV_DISABLE, item );
   ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );
}

void delete_timer( struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = 0;
   kev->flags = EV_DELETE;
   kev->udata = item;
}

void set_timer( struct ccb *item, int read )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = ( read ? read_timeout * 1000 : write_timeout * 60000 );
   kev->flags = EV_ADD;
   kev->udata = item;
}

void non_blocking( int desc )
{
   int flags, unblocked;

   if (( flags = fcntl( desc, F_GETFL, 0 )) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }

   unblocked = flags & O_NONBLOCK;

   if ( ! unblocked && fcntl( desc, F_SETFL, flags | O_NONBLOCK ) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }
}

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "dorrit: malloc(): %s\n", strerror( errno ));
   }

   return ptr;
}

struct stack *make_stack()
{
   struct stack *a;

   if (( a = ( struct stack *)memory( sizeof( struct stack ))) == NULL )
      return NULL;

   if (( a->values = memory( sizeof( union stack_u ) * increment )) == NULL )
      return NULL;

   a->free = increment;
   a->used = 0;
   a->top = a->values;

   return a;
}

void stack_free( struct stack *a )
{
   free( a->values );
   free( a );
}

int stack_push( struct stack *a, void *o )
{
   if ( a->free == 0 )
   {
      a->values = realloc( a->values, sizeof( union stack_u ) *
                                      ( a->used + increment ) );

      if ( a->values == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "dorrit: realloc(): %s\n", strerror( errno ));

         return 1;
      }

      a->free = increment;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   a->top->ptr = o;
   --a->free;
   ++a->used;

   return 0;
}

int stack_push_pid( struct stack *a, pid_t o )
{
   if ( a->free == 0 )
   {
      a->values = realloc( a->values, sizeof( union stack_u ) *
                                      ( a->used + increment ) );

      if ( a->values == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "dorrit: realloc(): %s\n", strerror( errno ));

         return 1;
      }

      a->free = increment;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   a->top->pid = o;
   --a->free;
   ++a->used;

   return 0;
}

char *stack_pop( struct stack *a )
{
   char *ptr;

   if ( a->used )
   {
      ptr = a->top->ptr;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return ptr;
   }

   return NULL;
}

pid_t stack_pop_pid( struct stack *a )
{
   pid_t i;

   if ( a->used )
   {
      i = a->top->pid;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return i;
   }

   return 0;
}

char *str_dup( char *str, int len )
{
   char *ptr;

   if (( ptr = ( char *)memory( len + 1 )) != NULL )
   {
      bcopy( str, ptr, len );
      ptr[ len ] = '\0';
   }

   return ptr;
}

struct string *make_string()
{
   struct string *s;

   s = ( struct string *)memory( sizeof( struct string ));
   if ( s == NULL )
      return s;

   s->str = ( char *)memory( increment + 1 );
   *s->str = '\0';
   s->free = increment;
   s->used = 0;
   s->top = s->str;

   return s;
}

void string_free( struct string *s )
{
   free( s->str );
   free( s );
}

void signal_handler( int signo )
{
   while( children->used )
      kill( stack_pop_pid( children ), SIGTERM );

   exit( 0 );
}

struct string *string_append( struct string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + increment );

      if ( s->str == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "dorrit: realloc(): %s\n", strerror( errno ));

         return NULL;
      }

      /* Leave room for end-of-string sentinel */

      s->free = increment;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';

   return s;
}

void set_types()
{
   char *ptr, *dup, buffer[ 512 ];
   int fd, count;
   struct string *item;

   if (( type_table = make_stack()) == NULL )
      exit( 1 );

   if (( item = make_string() ) == NULL )
      exit( 1 );

   if (( fd = open( TYPEFILE, O_RDONLY )) < 0 )
   {
      fprintf( stderr, "dorrit: could not open %s\n", TYPEFILE );
      exit( 1 );
   }

   *buffer = '\0';
   ptr = buffer;

   for( ; ; )
   {
      if ( ! *ptr )
      {
         if (( count = read( fd, buffer, sizeof( buffer ) - 1 )) < 0 )
         {
            fprintf( stderr, "dorrit: could not read from %s: %s\n",
                     TYPEFILE, strerror( errno ));
            exit( 1 );
         }
         else if ( ! count )
         {
            if ( item->used )
               goto BREAK;

            break;
         }

         buffer[ count ] = '\0';
         ptr = buffer;
      }

      for( ; *ptr; ++ptr )
         switch( *ptr )
         {
            case '\t':
               while( *ptr == '\t' )
                  ++ptr;
               goto BREAK;

            case '\n':
               while( *ptr == '\n' )
                  ++ptr;
               goto BREAK;

            case '\r':
               continue;

            default:
               if ( string_append( item, *ptr ) == NULL )
                  exit( 1 );
         }

      if ( ! *ptr )
         continue;

      if ( ! item->used )
      {
         fprintf( stderr, "dorrit: empty field in %s\n", TYPEFILE );
         exit( 1 );
      }

   BREAK:
      if (( dup = str_dup( item->str, item->used )) == NULL )
         exit( 1 );

      if ( stack_push( type_table, dup ))
         exit( 1 );

      item->free += item->used;
      item->used = 0;
      item->top = item->str;
   }

   close( fd );
   string_free( item );

   if ( ! type_table->used || type_table->used % 2 )
   {
      fprintf( stderr, "dorrit: empty field in %s\n", TYPEFILE );
      exit( 1 );
   }

   if ( stack_push( type_table, "" ))
      exit( 1 );
}

void set_options( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xcq:t:w:i:p:r:u:g:e:m:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'c':
            do_chroot = 1;
            break;

         case 'e':
            root_resource = optarg;
            break;

         case 'g':
            grp = optarg;
            break;

         case 'i':
            interface = optarg;
            break;

         case 'm':
            instances = atoi( optarg );
            break;

         case 'p':
            port = optarg;
            break;

         case 'r':
            root = optarg;
            break;

         case 't':
            read_timeout = atoi( optarg );
            break;

         case 'u':
            user = optarg;
            break;

         case 'q':
            backlog = atoi( optarg );
            break;

         case 'w':
            write_timeout = atoi( optarg );
            break;

         case 'x':
            testing = 1;
      }

   if ( root == NULL )
   {
      fputs( "dorrit: server root directory not specified\n", stderr );
      exit( 1 );
   }

   if ( instances < 1 )
   {
      fprintf( stderr, "dorrit: instances to run < 1: %d\n", instances );
      exit( 1 );
   }

   if ( read_timeout < 1 || read_timeout > 60 )
   {
      fprintf( stderr, "dorrit: read timeout value %d out of range 1-60\n",
               read_timeout );
      exit( 1 );
   }

   if ( write_timeout < 0 || write_timeout > 120 )
   {
      fprintf( stderr, "dorrit: write timeout value %d out of range 0-120\n",
               write_timeout );
      exit( 1 );
   }

   if (( passwd = getpwnam( user )) == NULL )
   {
      fprintf( stderr, "dorrit: user \"%s\" does not exist\n", user );
      exit( 1 );
   }

   if (( group = getgrnam( grp )) == NULL )
   {
      fprintf( stderr, "dorrit: group \"%s\" does not exist\n", grp );
      exit( 1 );
   }
}

void failed_rx( int result, regex_t *rx )
{
   char err[ 80 ];

   regerror( result, rx, err, sizeof( err ));
   fprintf( stderr, "dorrit: regcomp(): %s\n", err );
   exit( 1 );
}

void init_rx()
{
   int flags, result;

   flags = REG_EXTENDED | REG_ICASE;

   if (( result =
            regcomp( &request_rx,
               "^([^ \t:]+)[ \t]+([^ \t]+)[ \t]+(HTTP/[0-9]+\\.[0-9]+)$",
               flags )))
      failed_rx( result, &request_rx );

   if (( result = regcomp( &type_rx, "^Content-Type:[\t ]+([^ \t]+)",
                          flags )))
      failed_rx( result, &type_rx );

   if (( result = regcomp( &host_rx, "^Host:[ \t]+([^\t ]+)", flags )))
      failed_rx( result, &host_rx );

   if (( result = regcomp( &close_rx, "^Connection:[ \t]+close", flags )))
      failed_rx( result, &close_rx );

   if (( result = regcomp( &parent_rx, "\\.\\./", flags )))
      failed_rx( result, &parent_rx );

   if (( result = regcomp( &suffix_rx, "\\.([^. \t]+$)", flags )))
      failed_rx( result, &suffix_rx );

   if (( result = regcomp( &since_rx, "^If-Modified-Since:[\t ]+(.+)$",
        flags )))
      failed_rx( result, &since_rx );

   if (( result = regcomp( &unmodified_rx, "^If-Unmodified-Since:[\t ]+(.+)$",
        flags )))
      failed_rx( result, &unmodified_rx );

   if (( result = regcomp( &gzip_rx,
                           "^Accept-Encoding:.*[, \t]gzip(; q=([.0-9]+))?",
                           flags )))
      failed_rx( result, &gzip_rx );
}

void set_signals()
{
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * Setting SIGCHLD to SIG_IGN prevents zombies from occurring.  We don't
    * need to wait() for our children.
    */

   signal( SIGCHLD, SIG_IGN );

   /*
    * We ignore everything but SIGTERM when ! testing.  Otherwise, we
    * catch all the signals, so we may kill our children before exiting.
    */

   signal( SIGTERM, signal_handler );

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( testing ? signal_handler : SIG_IGN ));
}

void become_daemon()
{
   int file;

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "dorrit: fork(): %s\n", strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   if ( !testing )
   {
      void *ptr;
      mode_t mode;
      char buffer[ 16 ];

      /*
       * Our pid has changed after the fork() above.
       */

      snprintf( buffer, sizeof( buffer ), "%d", getpid() );

      if (( ptr = setmode( "0600" )) == NULL )
      {
         syslog( LOG_ERR, "setmode( 0600 ): %m" );
         return;
      }

      mode = getmode( ptr, 0 );
      free( ptr );

      if (( file = open( PIDFILE, O_WRONLY | O_CREAT | O_TRUNC, mode )) < 0 )
         syslog( LOG_WARNING, "open( %s ): %m", PIDFILE );
      else
      {
         write( file, buffer, strlen( buffer ));
         close( file );
      }
   }

   umask( 0 );
}

void change_identity()
{
   if ( chdir( root ) < 0 )
   {
      syslog( LOG_ERR, "chdir(): %m" );
      exit( 1 );
   }

   if ( do_chroot && chroot( root ) < 0 )
   {
      syslog( LOG_ERR, "chroot(): %m" );
      exit( 1 );
   }

   if ( setgid( group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "setgid(): %m" );
      exit( 1 );
   }

   if ( setuid( passwd->pw_uid ) < 0 )
   {
      syslog( LOG_ERR, "setuid(): %m" );
      exit( 1 );
   }
}

void start_listening()
{
   struct addrinfo hints, *res;
   struct accept_filter_arg afa;
   int result;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified a listening interface, we use
    * the IPv6 wildcard.
    */

   if (( result = getaddrinfo( ( *interface ? interface : NULL ), port,
                               &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   if ( setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   result = 0;

   /*
    * Note we are shutting off this option if we are binding to the IPv6
    * wildcard interface, so we may accept both IPv4 and IPv6 connections.
    */

   if ( ! *interface &&
        setsockopt( fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( fd, backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   /*
    * Accept filter must be installed after socket is listening.
    */

   bzero( &afa, sizeof( afa ));
   strcpy( afa.af_name, "httpready" );

   if ( setsockopt( fd, SOL_SOCKET, SO_ACCEPTFILTER, &afa,
                    sizeof( afa )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_ACCEPTFILTER ): %m" );
}

void add_conn( int new, char *addr )
{
   struct ccb *ptr;

   /*
    * Drop the connection if we have run out of memory.
    */

   if (( ptr = memory( sizeof( struct ccb ))) == NULL )
   {
      close( new );
      return;
   }

   if (( ptr->addr = str_dup( addr, strlen( addr ))) == NULL )
   {
      free( ptr );
      close( new );
      return;
   }

   *ptr->buffer = '\0';
   *ptr->inbuf = '\0';
   ptr->inptr = ptr->inbuf;

   ptr->close = ptr->count = ptr->error = ptr->inactive =
      ptr->gzip = ptr->timer = 0;

   ptr->size = ptr->written = 0;

   ptr->state = RECEIVE;

   ptr->verno = ptr->type = -1;
   ptr->sock = new;
   ptr->source = 0;

   ptr->host = ptr->version = ptr->agent =
      ptr->request = ptr->method = ptr->since = ptr->unmodified = NULL;

   ptr->request_found = ptr->close_found = ptr->gzip_found =
      ptr->host_found =
      ptr->since_found = ptr->unmodified_found = 0;

   ev_set( ptr->sock, EVFILT_READ, EV_ADD, ptr );
   ev_set( ptr->sock, EVFILT_WRITE, EV_ADD | EV_DISABLE, ptr );
   set_timer( ptr, 1 );
}

void clear_conn( struct ccb *item )
{
   if ( item->source )
   {
      close( item->source );
      item->source = 0;
   }

   if ( item->host != NULL )
   {
      free( item->host );
      item->host = NULL;
   }

   if ( item->version != NULL )
   {
      free( item->version );
      item->version = NULL;
   }

   if ( item->agent != NULL )
   {
      free( item->agent );
      item->agent = NULL;
   }

   if ( item->request != NULL )
   {
      free( item->request );
      item->request = NULL;
   }

   if ( item->method != NULL )
   {
      free( item->method );
      item->method = NULL;
   }

   if ( item->since != NULL )
   {
      free( item->since );
      item->since = NULL;
   }

   if ( item->unmodified != NULL )
   {
      free( item->unmodified );
      item->unmodified = NULL;
   }

   *item->buffer = '\0';

   item->close = item->count = item->error = item->gzip =
      item->inactive = item->timer = 0;

   item->request_found = item->close_found =
      item->host_found = item->gzip_found =
      item->since_found = item->unmodified_found = 0;

   item->state = RECEIVE;

   item->size = item->written = item->type = item->verno = -1;

   if ( item->sock >= 0 )
   {
      ev_set( item->sock, EVFILT_READ, EV_ENABLE, item );

      if ( ! *item->inptr )
         ev_set( item->sock, EVFILT_WRITE, EV_DISABLE , item );

      delete_timer( item );
      set_timer( item, 1 );
   }
}

void remove_conn( struct ccb *item )
{
   if ( item->sock < 0 )
      return;

   if ( item->addr != NULL )
   {
      free( item->addr );
      item->addr = NULL;
   }

   closed = item->sock;
   delete_timer( item );

   close( item->sock );
   item->sock = -1;

   clear_conn( item );

   free( item );
}

char **match( regex_t *regexp, char *the_string, int len, int subexpr )
{
   static struct string *buffer = NULL;
   static char *results[ 3 ];
   regmatch_t matches[ 4 ];
   int result, length, i, j, k;

   if ( buffer == NULL && ( buffer = make_string() ) == NULL )
      return NULL;

   matches[ 0 ].rm_so = 0;
   matches[ 0 ].rm_eo = len;

   result = regexec( regexp, the_string, ( subexpr ? 4 : 0 ), matches,
                     REG_STARTEND );

   if ( result )
   {
      char err[ 80 ];

      if ( result == REG_NOMATCH )
      {
         matched = 0;
         return NULL;
      }

      regerror( result, regexp, err, sizeof( err ));
      syslog( LOG_ERR, "regexec: %m" );
      exit( 1 );
   }

   matched = 1;
   results[ 0 ] = results[ 1 ] = results[ 2 ] = NULL;

   if ( ! subexpr )
      return results;

   for( k = 0, i = 1; i < 4; ++i, ++k )
   {
      buffer->free += buffer->used;
      buffer->used = 0;
      buffer->top = buffer->str;

      length = 0;

      if ( matches[ i ].rm_so >= 0 )
      {
         length = matches[ i ].rm_so + matches[ i ].rm_eo - matches[ i ].rm_so;

         for( j = matches[ i ].rm_so; j < length; ++j )
            if ( string_append( buffer, the_string[ j ] ) == NULL )
            {
               if ( k )
               {
                  free( results[ 0 ] );
                  if ( k == 2 )
                     free( results[ 1 ] );
               }

               return NULL;
            }
      }

      if ( matches[ i ].rm_so >= 0 && length )
      {
         if (( results[ k ] = str_dup( buffer->str, buffer->used ))
               == NULL )
         {
            if ( k )
            {
               free( results[ 0 ] );
               if ( k == 2 )
                  free( results[ 1 ] );
            }

            return NULL;
         }
      }
      else
         results[ k ] = NULL;
   }

   return results;
}

void receive_request( struct ccb *item )
{
   char *line, **matches;
   int r, notready;

   for( ; ; )
   {
      notready = 0;

      if ( ! item->inactive )
      {
         item->count = 0;
         *item->buffer = '\0';
      }

      for( ; ; )
      {
         if ( ! *item->inptr )
         {
            r = read( item->sock, item->inbuf, INBUFSIZE - 1 );

            if ( r <= 0 )
            {
               if ( r < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ))
                  notready = 1;

               line = NULL;
               break;
            }

            item->inbuf[ r ] = '\0';
            item->inptr = item->inbuf;
         }

         /*
          * We don't want the line terminators.
          */

         if ( *item->inptr == 13 )
         {
            ++item->inptr;
            continue;
         }

         if ( *item->inptr == 10 )
         {
            line = item->buffer;
            ++item->inptr;
            break;
         }

         /*
          * New character overwrites previous string terminator.  New string
          * terminator added.  Notice that item->count does not count the
          * string terminator.
          */

         item->buffer[ item->count++ ] = *item->inptr++;

         if ( item->count >= IOBUFSIZE )
         {
            notready = 2;
            line = NULL;
            break;
         }

         item->buffer[ item->count ] = '\0';
      }

      if ( line == NULL )
      {
         switch( notready )
         {
            /*
             * 0 means premature EOF.
             */

            case 0:
               remove_conn( item );
               return;

            /*
             * 1 means the connection has become inactive.
             */

            case 1:
               item->inactive = 1;
               return;

            /*
             * 2 means the client sent a too-long line.  If this happens,
             * there is most likely malicious intent, and the request header
             * is now incomplete, so we drop the connection after sending
             * the error message.
             */

            case 2:
               item->count = 0;
               item->written = 0;
               item->close = 1;
               item->error = ERR_BADREQUEST;
               item->state = ERROR;

               delete_timer( item );
               flip_events( item );
               return;
         }
      }

      item->inactive = 0;

      /*
       * If we have the request header separator line, we are done here.
       */

      if ( *line == '\0' )
      {
         delete_timer( item );
         flip_events( item );

         if ( item->request == NULL || item->method == NULL ||
              item->version == NULL )
         {
            item->count = 0;
            item->written = 0;
            item->state = ERROR;
            item->error = ERR_BADREQUEST;
            return;
         }

         item->count = 0;
         item->written = 0;

         analyze_request( item );
         return;
      }

      /*
       * By using flags to prevent unnecessary applications of regular expressions
       * we speed up the server slightly when under load, when it matters.
       */

      if ( !item->request_found )
      {
         if (( matches = match( &request_rx, line, item->count, 1 )) == NULL )
         {
            if ( matched )
            {
               remove_conn( item );
               return;
            }
         }
         else
         {
            item->request_found = 1;

            /*
             * These may not be NULL if we are reading a (possibly
             * maliciously) screwed-up request header with multiple request
             * lines.  We prevent memory leaks by freeing the current strings
             * before assigning the new.
             */

            if ( item->method != NULL )
               free( item->method );

            item->method = matches[ 0 ];

            if ( item->request != NULL )
               free( item->request );

            item->request = matches[ 1 ];

            if ( item->version != NULL )
               free( item->version );

            item->version = matches[ 2 ];
            continue;
         }
      }

      /*
       * We limit the number of regular expressions which get applied to
       * each header line by testing the first letter of the line and
       * checking flags which are sent when a particular header has been
       * already seen.
       */

      if ( *line == 'C' || *line == 'c' )
      {
         if ( !item->close_found )
         {
            if ( match( &close_rx, line, item->count, 0 ) != NULL )
            {
               item->close_found = 1;
               item->close = 1;
               continue;
            }
         }

         continue;
      }

      if ( *line == 'H' || *line == 'h' )
      {
         if ( !item->host_found )
         {
            if (( matches = match( &host_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->host_found = 1;

               /*
                * Prevent memory leak if there are multiple Host headers.
                */

               if ( item->host != NULL )
                  free( item->host );

               item->host = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'I' || *line == 'i' )
      {
         if ( !item->since_found )
         {
            if (( matches =  match( &since_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->since_found = 1;

               if ( item->since != NULL )
                  free( item->since );

               item->since = matches[ 0 ];
               continue;
            }
         }

         if ( !item->unmodified_found )
         {
            if (( matches =  match( &unmodified_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->unmodified_found = 1;

               if ( item->unmodified!= NULL )
                  free( item->unmodified );

               item->unmodified = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'A' || *line == 'a' )
      {
         if ( item->count > 7 &&
              ( line[ 7 ] == 'E' || line[ 7 ] == 'e' ) &&
              !item->gzip_found )
         {
            if (( matches = match( &gzip_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->gzip_found = 1;
               item->gzip = 1;

               if ( matches[ 1 ] != NULL )
               {
                  free( matches[ 1 ] );

                  if ( matches[ 2 ] != NULL )
                  {
                     if ( !strcmp( matches[ 2 ], "0" ) || !strcmp( matches[ 2 ], "0.0" ))
                        item->gzip = 0;

                     free( matches[ 2 ] );
                  }
               }
            }
         }
      }
   }
}

void add_common_headers( struct ccb *item, int need_date )
{
   static char buffer[ 64 ];
   static time_t t = 0, ot = 0;
   struct tm *lt;
   int len;

   strcat( item->buffer, server_header );

   if ( need_date )
   {
      if (( t = time( NULL )) != ot )
      {
         ot = t;
         lt = gmtime( &t );
         strftime( buffer, sizeof( buffer ),
                   "Date: %a, %d %b %Y %H:%M:%S %Z\r\n", lt );
         len = strlen( buffer ) - 2;

         buffer[ --len ] = 'T';
         buffer[ --len ] = 'M';
         buffer[ --len ] = 'G';
      }

      strcat( item->buffer, buffer );
   }
}

void analyze_request( struct ccb *item )
{
   static struct string *s = NULL;
   struct stat st;
   struct tm *lt;
   char *ptr, **matches, type[ 128 ], size[ 128 ], modified[ 128 ];
   int g = 1, h = 1, len;

   if ( ( item->verno = strcasecmp( item->version, "HTTP/1.0" )) &&
        strcasecmp( item->version, "HTTP/1.1" ))
   {
      item->error = ERR_BADVERSION;
      item->close = 1;
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      return;
   }

   if ( ! item->verno )
      item->close = 1;

   if ( item->host == NULL )
   {
      if (( item->host = str_dup( "default", 7 )) == NULL )
      {
         remove_conn( item );
         return;
      }
   }
   else if ( *item->host == '.' || strchr( item->host, '/' ) != NULL )
   {
      item->count = 0;
      item->written = 0;
      item->close = 1;
      item->error = ERR_BADREQUEST;
      item->state = ERROR;
      return;
   }

   if ( ( g = strcasecmp( item->method, "GET" )) &&
        ( h = strcasecmp( item->method, "HEAD" )))
   {
      item->error = ERR_BADMETHOD;
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      return;
   }

   item->type = ( g ? 0 : 1 );

   if ( !( *item->request ))
   {
      item->error = ERR_BADREQUEST;
      item->count = 0;
      item->written = 0;
      item->close = 1;
      item->state = ERROR;
      return;
   }

   len = strlen( item->request );

   if ( match( &parent_rx, item->request, len, 0 ) != NULL )
   {
      item->error = ERR_BADREQUEST;
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->close = 1;
      return;
   }

   if ( s == NULL && ( s = make_string()) == NULL )
   {
      remove_conn( item );
      return;
   }

   s->free += s->used;
   s->used = 0;
   s->top = s->str;

   if ( item->request[ len - 1 ] == '/' )
   {
      len = strlen( item->request ) + strlen( root_resource );

      if (( ptr = memory( len + 1 )) == NULL )
      {
         remove_conn( item );
         return;
      }

      strcpy( ptr, item->request );
      strcat( ptr, root_resource );

      free( item->request );
      item->request = ptr;
   }

   /*
    * We make sure there is at least one separator between the various
    * items of the path we are going to create in s, because the
    * command-line arguments may not have included them, and the resource
    * specified in the request may have not begun with a slash.  Multiple
    * separators are tolerated by the filesystem-related system calls.
    */

   if ( string_append( s, '.' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   if ( string_append( s, '/' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   for ( ptr = item->host; *ptr; ++ptr )
      if ( string_append( s, *ptr ) == NULL )
      {
         remove_conn( item );
         return;
      }

   if ( string_append( s, '/' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   for( ptr = item->request; *ptr; ++ptr )
      if ( string_append( s, *ptr ) == NULL )
      {
         remove_conn( item );
         return;
      }

   if ( item->gzip )
   {
      if ( string_append( s, '.' ) == NULL )
      {
         remove_conn( item );
         return;
      }

      if ( string_append( s, 'g' ) == NULL )
      {
         remove_conn( item );
         return;
      }

      if ( string_append( s, 'z' ) == NULL )
      {
         remove_conn( item );
         return;
      }
   }

NO_GZIP:
   if ( stat( s->str, &st ) < 0 )
   {
      if ( item->gzip )
      {
         s->top -= 3;
         s->free += 3;
         s->used -= 3;
         s->str[ s->used ] = '\0';

         item->gzip = 0;
         goto NO_GZIP;
      }

      if ( errno == ENOENT || errno == ENOTDIR )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_NOTFOUND;
         return;
      }

      if ( errno == EACCES )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;
      return;
   }

   if ( ! S_ISREG( st.st_mode ))
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   /*
    * Don't send any file which is executable.
    */

   if (( S_IXUSR & st.st_mode ) ||( S_IXGRP & st.st_mode ) ||
         S_IXOTH & st.st_mode )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   if ( ! st.st_size )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_NOCONTENT;
      return;
   }

   if ( item->type && ( item->source = open( s->str, O_RDONLY )) < 0 )
   {
      if ( errno == EACCES )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, "open(): %m" );
      return;
   }

   ptr = NULL;

   if ( item->gzip )
   {
      s->top -= 3;
      s->free += 3;
      s->used -= 3;
      s->str[ s->used ] = '\0';
   }

   if (( matches = match( &suffix_rx, s->str, s->used, 1 )) == NULL )
   {
      if ( matched )
      {
         remove_conn( item );
         return;
      }
   }
   else
   {
      char **tp;

      for( tp = ( char **)type_table->values; **tp; tp += 2 )
         if ( ! strcmp( *tp, *matches ))
         {
            ptr = *( tp + 1 );
            break;
         }

      free( *matches );
   }

   if ( ptr == NULL )
      ptr = default_type;

   lt = gmtime( &st.st_ctime );
   strftime( modified, sizeof( modified ),
             "Last-Modified: %a, %d %b %Y %H:%M:%S %Z", lt );
   len = strlen( modified );

   modified[ --len ] = 'T';
   modified[ --len ] = 'M';
   modified[ --len ] = 'G';

   if ( item->since != NULL && ! strcmp( item->since, &modified[ 15 ] ))
   {
      item->state = ERROR;
      item->error = ERR_NOTMODIFIED;
      item->count = 0;
      item->written = 0;
      return;
   }

   if ( item->unmodified != NULL && strcmp( item->unmodified, &modified[ 15 ] ))
   {
      item->state = ERROR;
      item->error = ERR_MODIFIED;
      item->count = 0;
      item->written = 0;
      return;
   }

   strcat( modified, "\r\n" );

   /*
    * st.st_size is off_t which we cast to long long unsigned, since
    * there is no printf escape for off_t.
    */

   snprintf( type, sizeof( type ), "Content-Type: %s\r\n", ptr );
   snprintf( size, sizeof( size ), "Content-Length: %llu\r\n",
             ( item->type ? ( long long unsigned int )st.st_size :
                            ( long long unsigned int )0 ));

   *item->buffer = '\0';
   strcpy( item->buffer, item->version );
   strcat( item->buffer, " 200 OK\r\n" );

   strcat( item->buffer, type );
   strcat( item->buffer, size );

   if ( item->gzip )
      strcat( item->buffer, "Content-Encoding: gzip\r\n" );

   strcat( item->buffer, modified );
   add_common_headers( item, 1 );
   strcat( item->buffer, "\r\n" );

   item->count = strlen( item->buffer );
   item->written = 0;
   item->size = st.st_size;
   item->state = HEADER;
}

void respond_static( struct ccb *item )
{
   off_t written = 0;

   if ( sendfile( item->source, item->sock, item->written,
                  MIN( item->size, IOBUFSIZE ), NULL,
                  &written, 0 ) < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
      {
         remove_conn( item );
         return;
      }
   }

   item->written += written;
   item->size -= written;

   if ( ! item->size )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

/*
 * respond_header() and respond_error() are triggered by the writeability
 * of a client socket.  Because we know the socket is writable, we know
 * write() will not generate EWOULDBLOCK, but it can return short counts.
 * If write() returns an error, we drop the connection.
 */

void respond_header( struct ccb *item )
{
   off_t written;

   if ( write_timeout && ! item->timer )
   {
      set_timer( item, 0 );
      item->timer = 1;
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
   {
      item->state = RESPOND;
      item->written = 0;

      if ( ! item->type )
         ( item->close ? remove_conn( item ) : clear_conn( item ) );
      else
         respond_static( item );
   }
}

void respond_error( struct ccb *item )
{
   off_t written;

   if ( write_timeout && ! item->timer )
   {
      set_timer( item, 0 );
      item->timer = 1;
   }

   if ( ! item->count )
   {
      char buffer[ 512 ], len_buffer[ 64 ];

      *item->buffer = '\0';

      if ( item->version == NULL )
      {
         strcpy( item->buffer, "HTTP/1.0" );
         item->close = 1;
      }
      else
         strcpy( item->buffer, item->version );

      strcat( item->buffer, errstrings[ item->error ] );
      add_common_headers( item, 1 );

      snprintf( buffer, sizeof( buffer ),
                "<html><head><title>Error</title></head><body>%s</body></html>",
                errstrings[ item->error ] );

      snprintf( len_buffer, sizeof( len_buffer ), "Content-Length: %u\r\n",
                ( unsigned int )strlen( buffer ));

      strcat( item->buffer, len_buffer );
      strcat( item->buffer, "\r\n" );
      strcat( item->buffer, buffer );

      item->count = strlen( item->buffer );
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

int accept_connection()
{
   struct sockaddr_storage addr;
   char address[ 128 ];
   const char *ptr = NULL;
   socklen_t len;
   int conn;

   len = sizeof( struct sockaddr_storage );
   conn = accept( fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         syslog( LOG_ERR, "accept(): %m" );

      return 1;
   }

   non_blocking( conn );

   if ( addr.ss_family == AF_INET6 )
      ptr = inet_ntop( AF_INET6, &(( struct sockaddr_in6 *)&addr )->sin6_addr,
                       address, sizeof( address ));
   else if ( addr.ss_family == AF_INET )
      ptr = inet_ntop( AF_INET, &(( struct sockaddr_in *)&addr )->sin_addr,
                       address, sizeof( address ));

   if ( ptr == NULL )
      strcpy( address, "unknown" );

   add_conn( conn, address );
   return 0;
}

void remove_events( int n)
{
   int c;

   for( c = n + 1; c < out; ++c )
      if ( outqueue[ c ].ident == closed )
         outqueue[ c ].ident = -1;
}

void drop_connection( int n )
{
   remove_events( n );
   remove_conn( outqueue[ n ].udata );
}

/*
 * If a timer fires on a descriptor and that descriptor becomes readable
 * after the last invocation of kevent(), we could get the both events in
 * the queue, on the next invocation.  If the read event was earlier in the
 * queue, and in processing the read event, it was decided to close the
 * connection, its ccb would be freed with a call to remove_conn().  When
 * the timer event was encountered a second freeing of the ccb would occur.
 * If the events were placed in the queue in the opposite order, then
 * processing the timer event would call remove_conn(), and then the read
 * event handler would be dispatched with a pointer to the freed ccb.
 * Both situations will crash the server.

 * Our solution is to modify the output queue every time a connection is
 * dropped during the processing of a read, write, or timer event.  We make
 * sure any other events for the same descriptor further along in the queue
 * are ignored by setting the ident member of the appropriate struct kevent
 * to -1.

 * kevent() returns -1 when there is no room to create an error event in
 * the output queue.  We treat this case the same as an error event in the
 * output queue, and just move on.  Errors are generated from duplicate
 * delete timer events in the input queue, and the occasional failure to
 * create a read or a write event on a connection unexpectedly dropped by
 * the client or server.  When kevent() does return -1, the output queue
 * may contain events, but we don't know how many, so we can't service
 * them.  If they are errors, we don't care.  If they are IO events, they
 * are level-triggered, so we will get those events again.  If they are
 * timer events, they will be generated again, for timers are periodic.
 * The worst that can happen is that a timeout will take longer than
 * otherwise to occur.  No events are lost.

 * Because we the listening socket can only appear once in the output queue,
 * the rate at which we accept new connections is inversely proportional to
 * the number of active connections, but we do try and accept up to 100 at
 * once, if they are ready to be accepted.
 */

void process_clients()
{
   int kq, n;

   if (( kq = kqueue()) < 0 )
   {
      syslog( LOG_ERR, "kqueue(): %m" );
      exit( 1 );
   }

   non_blocking( fd );
   ev_set( fd, EVFILT_READ, EV_ADD | EV_ENABLE, NULL );

   for( ; ; )
   {
      out = kevent( kq, inqueue, in, outqueue, QLEN, NULL );
      in = 0;

      if ( out <= 0 )
         continue;

      for( n = 0; n < out; ++n )
      {
         if ( outqueue[ n ].ident < 0 || outqueue[ n ].flags & EV_ERROR )
            continue;

         if ( outqueue[ n ].filter == EVFILT_TIMER )
            drop_connection( n );
         else if ( outqueue[ n ].ident == fd )
         {
            int i;

            for( i = 0; i < 100; ++i )
               if ( accept_connection() )
                  break;
         }
         else
         {
            struct ccb *item;
            closed = 0;

            item = ( struct ccb *)outqueue[ n ].udata;

            /*
             * I have inlined the handler for RESPOND for speed.
             */

            if ( item->state != RESPOND )
               item->state( item );
            else
            {
               off_t written = 0;

               if ( sendfile( item->source, item->sock, item->written,
                              MIN( item->size, IOBUFSIZE ), NULL,
                              &written, 0 ) < 0 )
               {
                  if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
                  {
                     remove_conn( item );
                     continue;
                  }
               }

               item->written += written;
               item->size -= written;

               if ( ! item->size )
                  ( item->close ? remove_conn( item ) : clear_conn( item ));
            }

            if ( closed )
               remove_events( n );
         }
      }
   }
}

void fork_children()
{
   pid_t pid;

   if (( children = make_stack()) == NULL )
      exit( 1 );

   while( --instances )
      switch(( pid = fork() ))
      {
         case -1:
            syslog( LOG_ERR, "fork(): %m" );
            /* fall through intended */

         case 0:
            return;

         default:
            if ( stack_push_pid( children, pid ))
               exit( 1 );
      }
}

int main( int argc, char **argv )
{
   init_rx();

   set_options( argc, argv );
   set_signals();
   set_types();

   openlog( "dorrit", LOG_PID, LOG_DAEMON );
   logging = 1;

   if ( !testing )
      become_daemon();

   start_listening();
   fork_children();

   change_identity();
   process_clients();

   return 0;
}
