/*
 * Drood Copyright (c) 2007, 2008, 2009, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Drood is a minimal, multiplexing, kqueue-based, partially-compliant
 * HTTP/1.1 server.  The only parts of HTTP/1.1 it implements are
 * persistent connections, the "chunked" transfer-encoding, the "gzip"
 * content-encoding, and name-based virtual hosting dictated by the value
 * of the "Host" request header.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <dirent.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>
#include <regex.h>
#include <string.h>
#include <ctype.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.50"
#define PIDFILE "/var/run/drood.pid"
#define TYPEFILE DATADIR "/types.tab"

/*
 * Size of the output queue used by kevent().
 */

#define QLEN 1024

/*
 * This is size of the IO buffers used in connection control blocks.
 * Modifying this value may require a corresponding modification to
 * the precision of the chunk header conversion performed in
 * respond_chunked().  32K is the default size of a socket send buffer
 * on i386 FreeBSD 6.x.  We try to move data in chunks of this size in an
 * attempt to "fill the pipe" of each connection.
 *
 * This should never be set to less than 2000.
 */

#define IOBUFSIZE 32768

/*
 * This is the size of the input buffers used to read request headers.  It
 * should not be altered.  This is the maximum amount we ever want to
 * over-read into the next request header.  If the next request is a POST,
 * we don't want to read into the next request's body.

 * The request line and a Content-Length header will be the minimum header
 * we accept for POSTS, therefore we are able to over-read 35 characters
 * into the next request header safely.  This assumes the client only
 * terminated lines with newlines and not carriage-return newline pairs,
 * just in case.  An over-read of 35 characters will occur when the first
 * buffered character is the final newline terminating the current request
 * header.  The rest of the buffer will be over-read characters from the
 * next request header.  Include the string terminator, and that's why
 * INBUFSIZE is 37 and not 35.

 * We use the "first" flag in ccbs to cause receive_request() to read the
 * request line character-by-character.  If the request method is not POST,
 * then we reset the flag and subsequent reads will be buffered.  If the
 * request is a POST, receive_request() continues to read characters
 * individually, so as not to read into the current request's body.  If
 * the -d option is not present on the command-line, we are not handling
 * (S)CGI requests, so in that case, flag is always 0.  This gives us a
 * little performance boost.

 * Buffering 36 characters reduces the number of read() system calls we
 * make for every request header.  Previously, we read each character
 * separately.  This resulted in a significant degradation in performance
 * compared to Dorrit.  Dorrit uses a 512 INBUFSIZE because it rejects POST
 * requests.
 */

#define INBUFSIZE 37

/*
 * A connection is in one of the following states at any one time.  There
 * are handlers for each state which are invoked when kevent() returns an
 * event on a connection.  A connection may have the handler for its
 * current state called more than once as part of the multiplexing process.
 * Eventually a handler will advance a connection to a succeeding state,
 * unless a system error occurs, or the connection times-out, when it will
 * be dropped.  Any connection in a state described as a terminal state,
 * will transition back to the RECEIVE state for a persistent connection,
 * otherwise the connection will be closed.

 * RECEIVE:  Receive request header.  Transitions to HEADER or ERROR.

 * HEADER:   Send response header.  Is terminal state for HEAD requests and
 *           (S)CGI redirects.  When a response body follows the header,
 *           transitions to RESPOND, NOCHUNK, CHUNK, or CONTINUE.

 * RESPOND:  Send response body for static resource,
 *           Invoked at least once for each IOBUFSIZE portion of requested
 *           resource.  Terminal state.

 * NOCHUNK:  Send a response body from a (S)CGI program which either
 *           provided a Content-Length header, or was invoked by an
 *           HTTP/1.0 request.  Invoked at least once for every IOBUFSIZE
 *           protion of the (S)CGI program's output.  Terminal state.

 * CHUNK:    Chunk a response body on-the-fly for an HTTP/1.1 (S)CGI
 *           request.  Invoked at least once for each IOBUFSIZE - 8
 *           portion of the (S)CGI program's output.  Terminal state.

 * CONTINUE: Continue processing a POST request which contained an
 *           "Expect: continue" header, after "100 Continue" has been sent.
 *           Transitions to FILTER or ERROR.

 * ERROR:    Send an error response header and body.  Terminal state.

 * FILTER:   Read and supplement a (S)CGI response header, as
 *           necessary.  Transitions to ERROR or HEADER.

 * SCGI_CONNECT:  Initiate non-blocking connect to SCGI server.
 *                Transitions to SGI_PREPARE or ERROR.

 * SCGI_PREPARE:  Verify non-blocking connect to SCGI server succeeded, and
 *                prepare the SCGI header for transmission.
 *                Transitions to SCGI_HEADER or ERROR.

 * SCGI_HEADER:   Send environment as netstring header to SCGI server.
 *                Transitions to SGI_BODY or FILTER.

 * SCGI_BODY:     Send client request body to SCGI server.  Transitions to
 *                FILTER.

 */

/*
 * connection control block.
 */

struct ccb
{
   int type, verno, error, close, expect, count, source,
      sock, inactive, timer, gzip, first;

   int request_found, close_found, expect_found, gzip_found,
      host_found, type_found, agent_found, since_found, length_found,
      unmodified_found, referrer_found, date_found, location_found;

   void ( *state )( struct ccb * ), ( *next )( struct ccb * );

   off_t size, written, scgi_len;

   char *request, *method, *agent, *cookie, *host, *length, *version, *addr,
      *since, *unmodified, *referer, *content, *scgi_host, *scgi_port;

   char buffer[ IOBUFSIZE ], inbuf[ INBUFSIZE ], *inptr;
   struct stack *working;
};

void reset_signals();

void receive_request( struct ccb * );
void analyze_request( struct ccb * );
void respond_static( struct ccb * );

void filter_cgi_header( struct ccb * );
void respond_chunked( struct ccb * );
void respond_unchunked( struct ccb * );

void respond_error( struct ccb * );
void respond_header( struct ccb * );

void prepare_dynamic( struct ccb * );
void connect_scgi( struct ccb * );
void prepare_scgi( struct ccb * );
void send_scgi_header( struct ccb * );
void send_scgi_body( struct ccb * );

void ( *RECEIVE  )( struct ccb * ) = receive_request;
void ( *RESPOND  )( struct ccb * ) = respond_static;

void ( *ERROR    )( struct ccb * ) = respond_error;
void ( *HEADER   )( struct ccb * ) = respond_header;

void ( *FILTER   )( struct ccb * ) = filter_cgi_header;
void ( *CHUNK    )( struct ccb * ) = respond_chunked;
void ( *NOCHUNK  )( struct ccb * ) = respond_unchunked;
void ( *CONTINUE )( struct ccb * ) = prepare_dynamic;

void ( *SCGI_CONNECT  )( struct ccb * ) = connect_scgi;
void ( *SCGI_PREPARE  )( struct ccb * ) = prepare_scgi;
void ( *SCGI_HEADER   )( struct ccb * ) = send_scgi_header;
void ( *SCGI_BODY     )( struct ccb * ) = send_scgi_body;

/*
 * These must start at 1.  0 means "no error".
 */

#define ERR_BADVERSION  1
#define ERR_BADREQUEST  2
#define ERR_FORBIDDEN   3
#define ERR_LENGTHREQ   4
#define ERR_BADMETHOD   5
#define ERR_NOTFOUND    6
#define ERR_INTERNAL    7
#define ERR_NOCONTENT   8
#define ERR_NOTMODIFIED 9
#define ERR_MODIFIED    10

/*
 * KQueue Input and output queues.  We can add a maximun of four events to
 * the input queue for every item in the output queue, while processing
 * the output queue, except when the output event is for the listening
 * socket.  We can accept up to 100 connections when the listening socket
 * has them ready to go.  Each added connection adds three events to the
 * input queue.
 */

#define INQ QLEN * 4 + 300
struct kevent inqueue[ INQ ];
struct kevent outqueue[ QLEN ];

int in = 0, out = 0, closed = 0, read_timeout = 20, write_timeout = 60,
   instances = 1, scgi = 0, logfd = -1, child = 0;

volatile int restart = 0;

extern char *optarg;
char *root_resource = "index.html", 
     *server_header = "Server: drood/" VERSION "\r\n",
     **g_argv = NULL;

int fd = -1, increment = 1024,
   matched = 0, logging = 0, testing = 0, backlog = 1000, dynalen = 0;

struct passwd *passwd;
struct group *group;

char *interface = "", *port = "80", *dynamic = NULL, *clean_dynamic = NULL,
   *grp = "nobody", *user = "nobody", *root = NULL,
   *default_type = "text/plain; charset=utf-8", *logfile = NULL;

struct stack *children = NULL, *scgi_servers = NULL;

char *errstrings[ 11 ] = { "",
     " 505 HTTP Version Not Supported\r\n",
     " 400 Bad Request\r\n",
     " 403 Forbidden\r\n",
     " 411 Length Required\r\n",
     " 405 Method Not Allowed\r\nAllow: GET, HEAD, POST\r\n",
     " 404 Not Found\r\n",
     " 500 Internal Server Error\r\n",
     " 204 No Content\r\n",
     " 304 Not Modified\r\n",
     " 402 Precondition Failed\r\n"
};

regex_t request_rx, length_rx, host_rx, agent_rx, cookie_rx, gzip_rx,
   close_rx, expect_rx, location_rx, date_rx, parent_rx, 
   suffix_rx, type_rx, since_rx, unmodified_rx, referer_rx;

struct stack *type_table = NULL;

/*
 * String which grows dynamically.
 */

struct string
{
   int free, used;
   char *top, *str;
};

/*
 * Stack which grows dynamically.
 */

union stack_u {
   void *ptr;
   pid_t pid;
};

struct stack
{
   int free, used;
   union stack_u *top, *values;
};

void ev_set( int desc, short filter, u_short flags, struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = desc;
   kev->filter = filter;
   kev->fflags = 0;
   kev->flags = flags;
   kev->udata = item;
}

void flip_events( struct ccb *item )
{
   if ( item->source >= 0 )
      ev_set( item->source, EVFILT_READ, EV_DISABLE, item );

   ev_set( item->sock, EVFILT_READ, EV_DISABLE, item );
   ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );
}

void delete_timer( struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = 0;
   kev->flags = EV_DELETE;
   kev->udata = item;
}

void set_timer( struct ccb *item, int read )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = ( read ? read_timeout * 1000 : write_timeout * 60000 );
   kev->flags = EV_ADD;
   kev->udata = item;
}

void non_blocking( int desc )
{
   int flags, unblocked;

   if (( flags = fcntl( desc, F_GETFL, 0 )) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }

   unblocked = flags & O_NONBLOCK;

   if ( ! unblocked && fcntl( desc, F_SETFL, flags | O_NONBLOCK ) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      exit( 1 );
   }
}

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "drood: malloc(): %s\n", strerror( errno ));
   }

   return ptr;
}

struct stack *make_stack()
{
   struct stack *a;

   if (( a = ( struct stack *)memory( sizeof( struct stack ))) == NULL )
      return NULL;

   if (( a->values = memory( sizeof( union stack_u ) * increment )) == NULL )
      return NULL;

   a->free = increment;
   a->used = 0;
   a->top = a->values;

   return a;
}

void stack_free( struct stack *a )
{
   free( a->values );
   free( a );
}

int stack_push( struct stack *a, void *o )
{
   if ( a->free == 0 )
   {
      a->values = realloc( a->values, sizeof( union stack_u ) *
                                      ( a->used + increment ) );

      if ( a->values == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "drood: realloc(): %s\n", strerror( errno ));

         return 1;
      }

      a->free = increment;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   a->top->ptr = o;
   --a->free;
   ++a->used;

   return 0;
}

int stack_push_pid( struct stack *a, pid_t o )
{
   if ( a->free == 0 )
   {
      a->values = realloc( a->values, sizeof( union stack_u ) *
                                      ( a->used + increment ) );

      if ( a->values == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "drood: realloc(): %s\n", strerror( errno ));

         return 1;
      }

      a->free = increment;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   a->top->pid = o;
   --a->free;
   ++a->used;

   return 0;
}

char *stack_pop( struct stack *a )
{
   char *ptr;

   if ( a->used )
   {
      ptr = a->top->ptr;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return ptr;
   }

   return NULL;
}

pid_t stack_pop_pid( struct stack *a )
{
   pid_t i;

   if ( a->used )
   {
      i = a->top->pid;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return i;
   }

   return 0;
}

char *str_dup( char *str, int len )
{
   char *ptr;

   if (( ptr = ( char *)memory( len + 1 )) != NULL )
   {
      bcopy( str, ptr, len );
      ptr[ len ] = '\0';
   }

   return ptr;
}

struct string *make_string()
{
   struct string *s;

   s = ( struct string *)memory( sizeof( struct string ));
   if ( s == NULL )
      return s;

   s->str = ( char *)memory( increment + 1 );
   *s->str = '\0';
   s->free = increment;
   s->used = 0;
   s->top = s->str;

   return s;
}

void string_free( struct string *s )
{
   free( s->str );
   free( s );
}

void signal_handler( int signo )
{
   while( children->used )
      kill( stack_pop_pid( children ), SIGTERM );

   exit( 0 );
}

void sighup_handler( int signo )
{
   fprintf( stderr, "sighup!\n" );

   while( children->used )
      kill( stack_pop_pid( children ), SIGTERM );

   restart = 1;
}

void restart_drood()
{
   execvp( g_argv[ 0 ], g_argv );
   syslog( LOG_CRIT, "execvp(): %m" );
   exit( 1 );
}

void close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( F_SETFD, FD_CLOEXEC ): %m" );
}

struct string *string_append( struct string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + increment );

      if ( s->str == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "drood: realloc(): %s\n", strerror( errno ));

         return NULL;
      }

      /* Leave room for end-of-string sentinel */

      s->free = increment;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';

   return s;
}

void set_types()
{
   char *ptr, *dup, buffer[ 512 ];
   int fd, count;
   struct string *item;

   if (( type_table = make_stack()) == NULL )
      exit( 1 );

   if (( item = make_string() ) == NULL )
      exit( 1 );

   if (( fd = open( TYPEFILE, O_RDONLY )) < 0 )
   {
      fprintf( stderr, "drood: could not open %s\n", TYPEFILE );
      exit( 1 );
   }

   *buffer = '\0';
   ptr = buffer;

   for( ; ; )
   {
      if ( ! *ptr )
      {
         if (( count = read( fd, buffer, sizeof( buffer ) - 1 )) < 0 )
         {
            fprintf( stderr, "drood: could not read from %s: %s\n",
                     TYPEFILE, strerror( errno ));
            exit( 1 );
         }
         else if ( ! count )
         {
            if ( item->used )
               goto BREAK;

            break;
         }

         buffer[ count ] = '\0';
         ptr = buffer;
      }

      for( ; *ptr; ++ptr )
         switch( *ptr )
         {
            case '\t':
               while( *ptr == '\t' )
                  ++ptr;
               goto BREAK;

            case '\n':
               while( *ptr == '\n' )
                  ++ptr;
               goto BREAK;

            case '\r':
               continue;

            default:
               if ( string_append( item, *ptr ) == NULL )
                  exit( 1 );
         }

      if ( ! *ptr )
         continue;

      if ( ! item->used )
      {
         fprintf( stderr, "drood: empty field in %s\n", TYPEFILE );
         exit( 1 );
      }

   BREAK:
      if (( dup = str_dup( item->str, item->used )) == NULL )
         exit( 1 );

      if ( stack_push( type_table, dup ))
         exit( 1 );

      item->free += item->used;
      item->used = 0;
      item->top = item->str;
   }

   close( fd );
   string_free( item );

   if ( ! type_table->used || type_table->used % 2 )
   {
      fprintf( stderr, "drood: empty field in %s\n", TYPEFILE );
      exit( 1 );
   }

   if ( stack_push( type_table, "" ))
      exit( 1 );
}

void add_scgi_server( char *optarg )
{
   char *host, *port, *suffix;

   host = strsep( &optarg, "|" );
   port = strsep( &optarg, "|" );
   suffix = optarg;

   if ( host == NULL || port == NULL || suffix == NULL )
   {
      fputs( "drood: incomplete -s option value\n", stderr );
      exit( 1 );
   }

   if ( scgi_servers == NULL && ( scgi_servers = make_stack()) == NULL )
      exit( 1 );

   if ( stack_push( scgi_servers, suffix ) ||
        stack_push( scgi_servers, host ) ||
        stack_push( scgi_servers, port ))
      exit( 1 );

   scgi = 1;
}

void set_options( int argc, char **argv )
{
   int i;

   g_argv = argv;

   while(( i = getopt( argc, argv, "xl:s:q:t:w:i:p:r:d:u:g:e:m:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'e':
            root_resource = ( optarg == NULL ? "index.html" : optarg );
            break;

         case 'g':
            grp = optarg;
            break;

         case 'i':
            interface = optarg;
            break;

         case 'm':
            instances = atoi( optarg );
            break;

         case 'l':
            logfile = optarg;
            break;

         case 'p':
            port = optarg;
            break;

         case 'r':
            root = optarg;
            break;

         case 's':
            add_scgi_server( optarg );
            break;

         case 't':
            read_timeout = atoi( optarg );
            break;

         case 'd':
            dynamic = optarg;
            break;

         case 'u':
            user = optarg;
            break;

         case 'q':
            backlog = atoi( optarg );
            break;

         case 'w':
            write_timeout = atoi( optarg );
            break;

         case 'x':
            testing = 1;
      }

   if ( root == NULL )
   {
      fputs( "drood: server root directory not specified\n", stderr );
      exit( 1 );
   }

   if ( instances < 1 )
   {
      fprintf( stderr, "drood: instances to run < 1: %d\n", instances );
      exit( 1 );
   }

   if ( dynamic != NULL )
   {
      struct string *s;
      char *ptr;

      clean_dynamic = dynamic;

      if ( ! strcmp( dynamic, ".." ) || ! strcmp( dynamic, "." ))
      {
         fprintf( stderr,
                  "drood: CGI directory name cannot be \".\" or \"..\"\n" );
         exit( 1 );
      }

      if (( s = make_string() ) == NULL )
         exit( 1 );

      if ( string_append( s, '/' ) == NULL )
         exit( 1 );

      for( ptr = dynamic; *ptr; ++ptr )
         if ( *ptr == '/' )
         {
            fprintf( stderr,
                     "drood: CGI directory name contains virgule(s)\n" );
            exit( 1 );
         }
         else if ( string_append( s, *ptr ) == NULL )
            exit( 1 );

      if ( string_append( s, '/' ) == NULL )
         exit( 1 );

      dynamic = s->str;
      dynalen = s->used;

      free( s ); /* not string_free() */
   }

   if ( read_timeout < 1 || read_timeout > 60 )
   {
      fprintf( stderr, "drood: read timeout value %d out of range 1-60\n",
               read_timeout );
      exit( 1 );
   }

   if ( write_timeout < 0 || write_timeout > 120 )
   {
      fprintf( stderr, "drood: write timeout value %d out of range 0-120\n",
               write_timeout );
      exit( 1 );
   }

   if (( passwd = getpwnam( user )) == NULL )
   {
      fprintf( stderr, "drood: user \"%s\" does not exist\n", user );
      exit( 1 );
   }

   if (( group = getgrnam( grp )) == NULL )
   {
      fprintf( stderr, "drood: group \"%s\" does not exist\n", grp );
      exit( 1 );
   }
}

void failed_rx( int result, regex_t *rx )
{
   char err[ 80 ];

   regerror( result, rx, err, sizeof( err ));
   fprintf( stderr, "drood: regcomp(): %s\n", err );
   exit( 1 );
}

void init_rx()
{
   int flags, result;

   flags = REG_EXTENDED | REG_ICASE;

   if (( result =
            regcomp( &request_rx,
               "^([^ \t:]+)[ \t]+([^ \t]+)[ \t]+(HTTP/[0-9]+\\.[0-9]+)$",
               flags )))
      failed_rx( result, &request_rx );

   if (( result = regcomp( &length_rx, "^Content-Length:[\t ]+([^ \t]+)",
                          flags )))
      failed_rx( result, &length_rx );

   if (( result = regcomp( &type_rx, "^Content-Type:[\t ]+([^ \t]+)",
                          flags )))
      failed_rx( result, &type_rx );

   if (( result = regcomp( &host_rx, "^Host:[ \t]+([^\t ]+)", flags )))
      failed_rx( result, &host_rx );

   if (( result = regcomp( &cookie_rx, "^Cookie:[ \t]+(.+)", flags )))
      failed_rx( result, &cookie_rx );

   if (( result = regcomp( &agent_rx, "^User-Agent:[ \t]+(.+)", flags )))
      failed_rx( result, &agent_rx );

   if (( result = regcomp( &close_rx, "^Connection:[ \t]+close", flags )))
      failed_rx( result, &close_rx );

   if (( result = regcomp( &expect_rx, "^Expect:[ \t]+100-continue", flags )))
      failed_rx( result, &expect_rx );

   if (( result = regcomp( &location_rx, "^Location:[ \t]+[^\t ]+",
                           flags )))
      failed_rx( result, &location_rx );

   if (( result = regcomp( &date_rx, "^Date:[ \t]+[^ \t]+", flags )))
      failed_rx( result, &date_rx );

   if (( result = regcomp( &parent_rx, "\\.\\./", flags )))
      failed_rx( result, &parent_rx );

   if (( result = regcomp( &suffix_rx, "\\.([^. \t]+$)", flags )))
      failed_rx( result, &suffix_rx );

   if (( result = regcomp( &since_rx, "^If-Modified-Since:[\t ]+(.+)$",
        flags )))
      failed_rx( result, &since_rx );

   if (( result = regcomp( &unmodified_rx, "^If-Unmodified-Since:[\t ]+(.+)$",
        flags )))
      failed_rx( result, &unmodified_rx );

   if (( result = regcomp( &referer_rx, "^Referr?er:[\t ]+(.+)$",
        flags )))
      failed_rx( result, &referer_rx );

   if (( result = regcomp( &gzip_rx,
                           "^Accept-Encoding:.*[, \t]gzip(; q=([.0-9]+))?",
                           flags )))
      failed_rx( result, &gzip_rx );
}

void reset_signals()
{
   int *iptr;
   int sigs[] = { SIGTERM, SIGHUP, SIGPIPE, SIGQUIT, SIGUSR1, SIGUSR2,
                  SIGALRM, SIGINT, SIGTSTP, -1 };

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, SIG_DFL );
}

void set_signals()
{
   struct sigaction sigact;
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   if ( testing )
      signal( SIGHUP, signal_handler );
   else
   {
      sigact.sa_handler = sighup_handler;
      sigemptyset( &sigact.sa_mask );
      sigact.sa_flags = 0;

      /*
       * SIGHUP is set to interrupt system calls.
       */

      if ( sigaction( SIGHUP, &sigact, NULL ) < 0 )
         fprintf( stderr, "drood: sigaction( SIGHUP ): %s.\n", strerror( errno ));
   }

   /*
    * Setting SIGCHLD to SIG_IGN prevents zombies from occurring.  We don't
    * need to wait() for our children.
    */

   signal( SIGCHLD, SIG_IGN );

   /*
    * We ignore everything but SIGTERM and SIGHUP when ! testing.
    * Otherwise, we catch all the signals, so we may kill our children
    * before exiting.
    */

   signal( SIGTERM, signal_handler );

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( testing ? signal_handler : SIG_IGN ));
}

void become_daemon()
{
   int file, count, restart = 0;
   char buffer[ 16 ], buffer2[ 16 ];
   void *ptr;
   mode_t mode;

   umask( 0 );
   snprintf( buffer2, sizeof( buffer2 ), "%d", getpid() );

   if (( file = open( PIDFILE, O_RDONLY )) > 0 )
   {
      if (( count = read( file, buffer, sizeof( buffer ) - 1 )) > 0 )
      {
         buffer[ count ] = '\0';

         /*
          * If the current pid is the same as the pid in PIDFILE,
          * then we are restarting in response to a SIGHUP.
          */

         if ( !strcmp( buffer, buffer2 ))
            restart = 1;
      }

      close( file );

      if ( restart )
         return;
   }

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "drood: fork(): %s\n", strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Our pid has changed after the fork() above.
    */

   snprintf( buffer2, sizeof( buffer2 ), "%d", getpid() );

   if (( ptr = setmode( "0600" )) == NULL )
   {
      syslog( LOG_ERR, "setmode( 0600 ): %m" );
      return;
   }

   mode = getmode( ptr, 0 );
   free( ptr );

   if (( file = open( PIDFILE, O_WRONLY | O_CREAT | O_TRUNC, mode )) < 0 )
      syslog( LOG_WARNING, "open( %s ): %m", PIDFILE );
   else
   {
      write( file, buffer2, strlen( buffer2 ));
      close( file );
   }
}

void change_identity()
{
   if ( setgid( group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "setgid(): %m" );
      exit( 1 );
   }

   if ( setuid( passwd->pw_uid ) < 0 )
   {
      syslog( LOG_ERR, "setuid(): %m" );
      exit( 1 );
   }
}

void start_listening()
{
   struct addrinfo hints, *res;
   struct accept_filter_arg afa;
   int result;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified a listening interface, we use
    * the IPv6 wildcard.
    */

   if (( result = getaddrinfo( ( *interface ? interface : NULL ), port,
                               &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   if ( setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   result = 0;

   /*
    * Note we are shutting off this option if we are binding to the IPv6
    * wildcard interface, so we may accept both IPv4 and IPv6 connections.
    */

   if ( ! *interface &&
        setsockopt( fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( fd, backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   /*
    * Accept filter must be installed after socket is listening.
    */

   bzero( &afa, sizeof( afa ));
   strcpy( afa.af_name, "httpready" );

   if ( setsockopt( fd, SOL_SOCKET, SO_ACCEPTFILTER, &afa,
                    sizeof( afa )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_ACCEPTFILTER ): %m" );
}

void add_conn( int new, char *addr )
{
   struct ccb *ptr;

   close_on_exec( new );

   /*
    * Drop the connection if we have run out of memory.
    */

   if (( ptr = memory( sizeof( struct ccb ))) == NULL )
   {
      close( new );
      return;
   }

   if (( ptr->addr = str_dup( addr, strlen( addr ))) == NULL )
   {
      free( ptr );
      close( new );
      return;
   }

   ptr->working = NULL;
   *ptr->buffer = '\0';
   *ptr->inbuf = '\0';
   ptr->inptr = ptr->inbuf;

   ptr->close = ptr->expect = ptr->count = ptr->error = ptr->inactive =
      ptr->gzip = ptr->timer = 0;

   ptr->first = ( dynamic != NULL ? 1 : 0 );

   ptr->size = ptr->written = ptr->scgi_len = 0;

   ptr->state = RECEIVE;
   ptr->next = NULL;

   ptr->source = ptr->verno = ptr->type = -1;
   ptr->sock = new;

   ptr->host = ptr->length = ptr->version = ptr->agent = ptr->cookie =
      ptr->request = ptr->method = ptr->since = ptr->unmodified =
      ptr->referer = ptr->content = ptr->scgi_host = ptr->scgi_port = NULL;

   ptr->request_found = ptr->close_found = ptr->expect_found = ptr->gzip_found =
      ptr->host_found = ptr->type_found = ptr->agent_found = ptr->length_found =
      ptr->since_found = ptr->unmodified_found = ptr->referrer_found =
      ptr->date_found = ptr->location_found = 0;

   ev_set( ptr->sock, EVFILT_READ, EV_ADD, ptr );
   ev_set( ptr->sock, EVFILT_WRITE, EV_ADD | EV_DISABLE, ptr );
   set_timer( ptr, 1 );
}

void clear_conn( struct ccb *item )
{
   char *ptr;

   if ( item->source >= 0 )
   {
      close( item->source );
      item->source = -1;
   }

   if ( item->host != NULL )
   {
      free( item->host );
      item->host = NULL;
   }

   if ( item->length != NULL )
   {
      free( item->length );
      item->length = NULL;
   }

   if ( item->version != NULL )
   {
      free( item->version );
      item->version = NULL;
   }

   if ( item->agent != NULL )
   {
      free( item->agent );
      item->agent = NULL;
   }

   if ( item->cookie != NULL )
   {
      free( item->cookie );
      item->cookie = NULL;
   }

   if ( item->request != NULL )
   {
      free( item->request );
      item->request = NULL;
   }

   if ( item->method != NULL )
   {
      free( item->method );
      item->method = NULL;
   }

   if ( item->since != NULL )
   {
      free( item->since );
      item->since = NULL;
   }

   if ( item->unmodified != NULL )
   {
      free( item->unmodified );
      item->unmodified = NULL;
   }

   if ( item->referer != NULL )
   {
      free( item->referer );
      item->referer = NULL;
   }

   if ( item->content != NULL )
   {
      free( item->content );
      item->content = NULL;
   }

   if ( item->scgi_host != NULL )
   {
      /*
       * Do not free!  Pointing to element in scgi_servers.
      */

      item->scgi_host = NULL;
   }

   if ( item->scgi_port != NULL )
   {
      /*
       * Do not free!  Pointing to element in scgi_servers.
       */

      item->scgi_port = NULL;
   }

   if ( item->working != NULL )
   {
      while(( ptr = ( char *)stack_pop( item->working )) != NULL )
         free( ptr );
   }

   item->first = ( dynamic != NULL ? 1 : 0 );

   *item->buffer = '\0';
   item->close = item->count = item->error = item->expect = item->gzip =
      item->inactive = item->timer = 0;

   item->request_found = item->close_found = item->expect_found =
      item->host_found = item->type_found = item->gzip_found =
      item->agent_found = item->length_found = item->since_found =
      item->unmodified_found = item->referrer_found =
      item->date_found = item->location_found = 0;

   item->state = RECEIVE;
   item->next = NULL;

   item->size = item->written = item->scgi_len = 0;
   item->type = item->verno = -1;

   if ( item->sock >= 0 )
   {
      ev_set( item->sock, EVFILT_READ, EV_ENABLE, item );

      /*
       * If there's still data in the input buffer, we have another
       * request to service.  The whole request may be in the buffer,
       * so we cannot rely on the socket becoming readable again to
       * trigger receive_request().  Write events on the socket
       * will have been disabled if the last request was an (S)CGI
       * request, so we we enable them.  When the socket is
       * writeable, receive_request() will get called.

       * Otherwise, if the input buffer is empty, we shut off write
       * events for the socket, and we wait for read events to trigger
       * receive_request().
       */

      if ( !*item->inptr )
         ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
      else
         ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );

      delete_timer( item );
      set_timer( item, 1 );
   }
}

void remove_conn( struct ccb *item )
{
   if ( item->sock < 0 )
      return;

   if ( item->addr != NULL )
   {
      free( item->addr );
      item->addr = NULL;
   }

   closed = item->sock;
   delete_timer( item );

   close( item->sock );
   item->sock = -1;

   clear_conn( item );

   if ( item->working != NULL )
      stack_free( item->working );

   free( item );
}

char **match( regex_t *regexp, char *the_string, int len, int subexpr )
{
   static struct string *buffer = NULL;
   static char *results[ 3 ];
   regmatch_t matches[ 4 ];
   int result, length, i, j, k;

   if ( buffer == NULL && ( buffer = make_string() ) == NULL )
      return NULL;

   matches[ 0 ].rm_so = 0;
   matches[ 0 ].rm_eo = len;

   result = regexec( regexp, the_string, ( subexpr ? 4 : 0 ), matches,
                     REG_STARTEND );

   if ( result )
   {
      char err[ 80 ];

      if ( result == REG_NOMATCH )
      {
         matched = 0;
         return NULL;
      }

      regerror( result, regexp, err, sizeof( err ));
      syslog( LOG_ERR, "regexec: %m" );
      exit( 1 );
   }

   matched = 1;
   results[ 0 ] = results[ 1 ] = results[ 2 ] = NULL;

   if ( ! subexpr )
      return results;

   for( k = 0, i = 1; i < 4; ++i, ++k )
   {
      buffer->free += buffer->used;
      buffer->used = 0;
      buffer->top = buffer->str;

      length = 0;

      if ( matches[ i ].rm_so >= 0 )
      {
         length = matches[ i ].rm_so + matches[ i ].rm_eo - matches[ i ].rm_so;

         for( j = matches[ i ].rm_so; j < length; ++j )
            if ( string_append( buffer, the_string[ j ] ) == NULL )
            {
               if ( k )
               {
                  free( results[ 0 ] );
                  if ( k == 2 )
                     free( results[ 1 ] );
               }

               return NULL;
            }
      }

      if ( matches[ i ].rm_so >= 0 && length )
      {
         if (( results[ k ] = str_dup( buffer->str, buffer->used ))
               == NULL )
         {
            if ( k )
            {
               free( results[ 0 ] );
               if ( k == 2 )
                  free( results[ 1 ] );
            }

            return NULL;
         }
      }
      else
         results[ k ] = NULL;
   }

   return results;
}

void receive_request( struct ccb *item )
{
   char *line, **matches;
   int r, notready;

   for( ; ; )
   {
      notready = 0;

      if ( ! item->inactive )
      {
         item->count = 0;
         *item->buffer = '\0';
      }

      for( ; ; )
      {
         if ( !*item->inptr )
         {
            r = read( item->sock, item->inbuf, ( item->first ? 1 : INBUFSIZE - 1 ));

            if ( r <= 0 )
            {
               if ( r < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ))
                  notready = 1;

               line = NULL;
               break;
            }

            item->inbuf[ r ] = '\0';
            item->inptr = item->inbuf;
         }

         /*
          * We don't want the line terminators.
          */

         if ( *item->inptr == 13 )
         {
            ++item->inptr;
            continue;
         }

         if ( *item->inptr == 10 )
         {
            ++item->inptr;
            line = item->buffer;
            break;
         }

         /*
          * New character overwrites previous string terminator.  New string
          * terminator added.  Notice that item->count does not count the
          * string terminator.
          */

         item->buffer[ item->count++ ] = *item->inptr++;

         if ( item->count >= IOBUFSIZE )
         {
            notready = 2;
            line = NULL;
            break;
         }

         item->buffer[ item->count ] = '\0';
      }

      if ( line == NULL )
      {
         switch( notready )
         {
            /*
             * 0 means premature EOF.
             */

            case 0:
               remove_conn( item );
               return;

            /*
             * 1 means the connection has become inactive.
             */

            case 1:
               item->inactive = 1;
               return;

            /*
             * 2 means the client sent a too-long line.  If this happens,
             * there is most likely malicious intent, and the request header
             * is now incomplete, so we drop the connection after sending
             * the error message.
             */

            case 2:
               item->count = 0;
               item->written = 0;
               item->close = 1;
               item->error = ERR_BADREQUEST;
               item->state = ERROR;

               delete_timer( item );
               flip_events( item );
               return;
         }
      }

      item->inactive = 0;

      /*
       * If we have the request header separator line, we are done here.
       */

      if ( *line == '\0' )
      {
         delete_timer( item );
         flip_events( item );

         if ( item->request == NULL || item->method == NULL ||
              item->version == NULL )
         {
            item->count = 0;
            item->written = 0;
            item->state = ERROR;
            item->error = ERR_BADREQUEST;
            return;
         }

         item->count = 0;
         item->written = 0;
         analyze_request( item );

         return;
      }

      /*
       * By using flags to prevent unnecessary applications of regular expressions
       * we speed up the server slightly when under load, when it matters.
       */

      if ( !item->request_found )
      {
         if (( matches = match( &request_rx, line, item->count, 1 )) == NULL )
         {
            if ( matched )
            {
               remove_conn( item );
               return;
            }
         }
         else
         {
            item->request_found = 1;

            /*
             * These may not be NULL if we are reading a (possibly
             * maliciously) screwed-up request header with multiple request
             * lines.  We prevent memory leaks by freeing the current strings
             * before assigning the new.
             */

            if ( item->method != NULL )
               free( item->method );

            item->method = matches[ 0 ];

            if (( dynamic != NULL ) && strcasecmp( item->method, "POST" ))
               item->first = 0;

            if ( item->request != NULL )
               free( item->request );

            item->request = matches[ 1 ];

            if ( item->version != NULL )
               free( item->version );

            item->version = matches[ 2 ];
            continue;
         }
      }

      /*
       * We limit the number of regular expressions which get applied to
       * each header line by testing the first letter of the line and
       * checking flags which are sent when a particular header has been
       * already seen.
       */

      if ( *line == 'C' || *line == 'c' )
      {
         if ( !item->close_found )
         {
            if ( match( &close_rx, line, item->count, 0 ) != NULL )
            {
               item->close_found = 1;
               item->close = 1;
               continue;
            }
         }

         if ( !item->type_found )
         {
            if (( matches = match( &type_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->type_found = 1;

               /*
                * Prevent memory leak if there are multiple Content-Type headers.
                */

               if ( item->content != NULL )
                  free( item->content );

               item->content = matches[ 0 ];
               continue;
            }
         }

         if ( !item->length_found )
         {
            if (( matches = match( &length_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->length_found = 1;

               /*
                * Prevent memory leak.
                */

               if ( item->length != NULL )
                  free( item->length );

               item->length = matches[ 0 ];
               continue;
            }
         }

         /*
          * We do not set a flag for the Cookie header, since there can be
          * more than one, legally.
          */

         if (( matches = match( &cookie_rx, line, item->count, 1 )) == NULL )
         {
            if ( matched )
            {
               remove_conn( item );
               return;
            }
         }
         else
         {
            /*
             * Combine multiple cookies.
             */

            if ( item->cookie != NULL )
            {
               int len, len2;

               if (( line = memory( ( len  = strlen( item->cookie )) +
                                    ( len2 = strlen( matches[ 0 ] )) +
                                    3 )) == NULL )
               {
                  remove_conn( item );
                  return;
               }

               memcpy( line, item->cookie, len );
               memcpy( &line[ len ], "; ", 2 );
               memcpy( &line[ len + 2 ], matches[ 0 ], len2 );
               line[ len + len2 + 2 ] = '\0';

               free( item->cookie );
               item->cookie = line;
            }
            else
               item->cookie = matches[ 0 ];
         }

         continue;
      }

      if ( *line == 'H' || *line == 'h' )
      {
         if ( !item->host_found )
         {
            if (( matches = match( &host_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->host_found = 1;

               /*
                * Prevent memory leak if there are multiple Host headers.
                */

               if ( item->host != NULL )
                  free( item->host );

               item->host = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'U' || *line == 'u' )
      {
         if ( !item->agent_found )
         {
            if (( matches = match( &agent_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->agent_found = 1;

               /*
                * Prevent memory leak.
                */

               if ( item->agent != NULL )
                  free( item->agent );

               item->agent = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'I' || *line == 'i' )
      {
         if ( !item->since_found )
         {
            if (( matches =  match( &since_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->since_found = 1;

               if ( item->since != NULL )
                  free( item->since );

               item->since = matches[ 0 ];
               continue;
            }
         }

         if ( !item->unmodified_found )
         {
            if (( matches =  match( &unmodified_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->unmodified_found = 1;

               if ( item->unmodified!= NULL )
                  free( item->unmodified );

               item->unmodified = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'E' || *line == 'e' )
      {
         if ( !item->expect_found )
         {
            if ( match( &expect_rx, line, item->count, 0 ) != NULL )
            {
               item->expect_found = 1;
               item->expect = 1;
            }
         }

         continue;
      }

      if ( *line == 'R' || *line == 'r' )
      {
         if ( !item->referrer_found )
         {
            if (( matches = match( &referer_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->referrer_found = 1;

               /*
                * Prevent memory leak.
                */

               if ( item->referer != NULL )
                  free( item->referer );

               item->referer = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'A' || *line == 'a' )
      {
         if ( item->count > 7 &&
              ( line[ 7 ] == 'E' || line[ 7 ] == 'e' ) &&
              !item->gzip_found )
         {
            if (( matches = match( &gzip_rx, line, item->count, 1 )) == NULL )
            {
               if ( matched )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->gzip_found = 1;
               item->gzip = 1;

               if ( matches[ 1 ] != NULL )
               {
                  free( matches[ 1 ] );

                  if ( matches[ 2 ] != NULL )
                  {
                     if ( !strcmp( matches[ 2 ], "0" ) || !strcmp( matches[ 2 ], "0.0" ))
                        item->gzip = 0;

                     free( matches[ 2 ] );
                  }
               }
            }
         }
      }
   }
}

void add_common_headers( struct ccb *item, int need_date )
{
   static char buffer[ 64 ];
   static time_t t = 0, ot = 0;
   struct tm *lt;
   int len;

   strcat( item->buffer, server_header );

   if ( need_date )
   {
      if (( t = time( NULL )) != ot )
      {
         ot = t;
         lt = gmtime( &t );
         strftime( buffer, sizeof( buffer ),
                   "Date: %a, %d %b %Y %H:%M:%S %Z\r\n", lt );
         len = strlen( buffer ) - 2;

         buffer[ --len ] = 'T';
         buffer[ --len ] = 'M';
         buffer[ --len ] = 'G';
      }

      strcat( item->buffer, buffer );
   }
}

int fork_cgi( char *prog, struct ccb *item )
{
   int pfd[ 2 ], flags, pid, ourend;
   char *ptr, *ptr2, *args[ 2 ];

   ourend = -1;

   if ( pipe( pfd ) < 0 )
   {
      syslog( LOG_ERR, "pipe(): %m" );
      return ourend;
   }

   switch(( pid = fork() ))
   {
      case -1:
         close( pfd[ 0 ] );
         close( pfd[ 1 ] );
         syslog( LOG_ERR, "fork(): %m" );
         return ourend;

      case 0:
         if ( ! child )
            change_identity();

         reset_signals();
         fclose( stdout );

         if (( dup2( pfd[ 1 ], 1 )) < 0 )
         {
            syslog( LOG_ERR, "dup2: %m" );
            _exit( 1 );
         }

         if (( stdout = fdopen( 1, "w" )) == NULL )
         {
            syslog( LOG_ERR, "fdopen(): %m" );
            _exit( 1 );
         }

         close( pfd[ 0 ] );
         close( pfd[ 1 ] );

         if ( item->type == 2 )
         {
            fclose( stdin );

            if (( dup2( item->sock, 0 )) < 0 )
            {
               syslog( LOG_ERR, "dup2: %m" );
               _exit( 1 );
            }

            if (( flags = fcntl( 0, F_GETFL, 0 )) < 0 )
            {
               syslog( LOG_ERR, "fcntl(): %m" );
               _exit( 1 );
            }

            if ( fcntl( 0, F_SETFL, flags & ~O_NONBLOCK ) < 0 )
            {
               syslog( LOG_ERR, "fcntl(): %m" );
               _exit( 1 );
            }

            if (( stdin = fdopen( 0, "r" )) == NULL )
            {
               syslog( LOG_ERR, "fdopen(): %m" );
               _exit( 1 );
            }
         }

         close( item->sock );

         while(( ptr = stack_pop( item->working )) != NULL )
         {
            ptr2 = stack_pop( item->working );
            setenv( ptr2, ptr, 1 );
            free( ptr );
            free( ptr2 );
         }

         args[ 0 ] = prog;
         args[ 1 ] = NULL;

         execv( args[ 0 ], args );
         syslog( LOG_ERR, "execv(): %m" );
         _exit( 1 );

      default:
         close( pfd[ 1 ] );
         ourend = pfd[ 0 ];
         non_blocking( ourend );
   }

   return ourend;
}

void send_scgi_body( struct ccb *item )
{
   off_t written;

   if ( ! item->count )
   {
      item->count = read( item->sock, item->buffer,
                          MIN( item->scgi_len, sizeof( item->buffer )));

      if ( item->count < 0 )
      {
         remove_conn( item );
         return;
      }

      /*
       * Since we do not read more than item->scgi_len chars, this
       * can only happen if the client closed (at least) the readable
       * half of the connection, prematurely.  This means the request
       * is borked, so we bail.
       */

      if ( ! item->count )
      {
         remove_conn( item );
         return;
      }
   }

   written = write( item->source, &item->buffer[ item->written ],
                    item->count );

   if ( written < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         remove_conn( item );

      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
   {
      item->scgi_len -= item->written;
      item->written = 0;

      if ( ! item->scgi_len )
      {
         *item->buffer = '\0';

         ev_set( item->source, EVFILT_READ, EV_ADD, item );
         ev_set( item->sock, EVFILT_READ, EV_DISABLE, item );

         item->type_found = item->date_found = item->location_found =
            item->length_found = 0;
         item->size = 0;
         item->state = FILTER;
      }
   }
}

void send_scgi_header( struct ccb *item )
{
   off_t written;

   written = write( item->source, &item->buffer[ item->written ],
                    item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
   {
      item->written = 0;
      ev_set( item->source, EVFILT_WRITE, EV_DISABLE, item );

      if ( item->scgi_len )
      {
         item->state = SCGI_BODY;
         ev_set( item->sock, EVFILT_READ, EV_ENABLE, item );
      }
      else
      {
         *item->buffer = '\0';
         item->type_found = item->date_found = item->location_found =
            item->length_found = 0;

         item->size = 0;
         item->state = FILTER;
         ev_set( item->source, EVFILT_READ, EV_ADD, item );
      }
   }
}

void prepare_scgi( struct ccb *item )
{
   char *var, *val, *ptr;
   int i, err, len, ns_len, total;
   socklen_t slen;

   slen = sizeof( err );
   err = 0;

   if ( getsockopt( item->source, SOL_SOCKET, SO_ERROR, &err, &slen ) < 0 )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, "getsockopt(): %m" );
      return;
   }

   if ( err )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, "connect(): %s", strerror( err ));
      return;
   }

   for( len = 0, i = 0; i < item->working->used; ++i )
      len += strlen( ( char *)item->working->values[ i ].ptr ) + 1;

   /*
    * Add seven characters for the SCGI header, added later.
    */

   len += 7;

   snprintf( item->buffer, sizeof( item->buffer ), "%d:", len );
   ns_len = strlen( item->buffer );

   /*
    * The extra character is for the comma,
    * added to the buffer after the CGI environment netstring.
    */

   if ( len + ns_len + 1 > IOBUFSIZE )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_BADREQUEST;

      return;
   }

   /*
    * CONTENT_LENGTH was last item pushed onto working stack, so it
    * will be the first item popped off.  It must come first in the header
    * according to the SCGI protocol.
    */

   item->scgi_len = ( off_t )atol( *( char **)item->working->top );
   len = ns_len;

   while(( val = stack_pop( item->working )) != NULL )
   {
      var = stack_pop( item->working );
      ptr = var;

      for( i = len, total = strlen( var ) + 1; total; ++i, --total )
      {
         item->buffer[ i ] = *ptr++;
         ++len;
      }

      ptr = val;

      for( i = len, total = strlen( val ) + 1; total; ++i, --total )
      {
         item->buffer[ i ] = *ptr++;
         ++len;
      }

      free( var );
      free( val );
   }

   ptr = "SCGI";
   for( i = 0; i < 5; ++i )
      item->buffer[ len++ ] = *ptr++;

   item->buffer[ len++ ] = '1';
   item->buffer[ len++ ] = '\0';
   item->buffer[ len++ ] = ',';

   item->count = len;
   item->written = 0;
   item->state = SCGI_HEADER;
}

void connect_scgi( struct ccb *item )
{
   int fd = -1, result;
   char buffer[ 255 ];
   struct addrinfo hints, *res, *ptr;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_family = PF_UNSPEC;
   hints.ai_socktype = SOCK_STREAM;

   if (( result = getaddrinfo( item->scgi_host, item->scgi_port, &hints, &res )))
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      snprintf( buffer, sizeof( buffer ), "getaddrinfo(): %s",
                gai_strerror( result ));
      syslog( LOG_ERR, buffer );
      return;
   }

   for( ptr = res; ptr != NULL; ptr = ptr->ai_next )
   {
      result = 0;
      fd = socket( ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol );

      if ( fd == -1 )
         continue;

      non_blocking( fd );

      if (( result = connect( fd, ptr->ai_addr, ptr->ai_addrlen )) == -1 )
      {
         if ( errno != EINPROGRESS )
         {
            close( fd );
            fd = -1;
            continue;
         }
      }

      break;
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( fd == -1 )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, ( result ? "connect(): %m" : "socket(): %m" ));
      return;
   }

   item->source = fd;
   item->state = SCGI_PREPARE;

   ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
   ev_set( item->source, EVFILT_WRITE, EV_ADD, item );
}

int add_to_env( struct ccb *item, char *str )
{
   char *ptr;

   if (( ptr = str_dup( str, strlen( str ))) == NULL )
      return 1;

   if ( stack_push( item->working, ptr ))
   {
      free( ptr );
      return 1;
   }

   return 0;
}

int find_scgi_server( struct ccb *item, char *prog )
{
   char **matches;
   int i;

   if (( matches = match( &suffix_rx, prog, strlen( prog ), 1 )) == NULL )
      return 0;

   for( i = 0; i < scgi_servers->used; i += 3 )
      if( !strcmp( scgi_servers->values[ i ].ptr, *matches ))
      {
         item->scgi_host = scgi_servers->values[ i + 1 ].ptr;
         item->scgi_port = scgi_servers->values[ i + 2 ].ptr;

         free( *matches );
         return 1;
      }

   item->scgi_host = NULL;
   item->scgi_port = NULL;
   free( *matches );

   return 0;
}

void prepare_dynamic( struct ccb *item )
{
   char *ptr, *prog = NULL, *args;
   struct stat st;
   static struct string *s = NULL;

   if ( item->expect )
   {
      *item->buffer = '\0';
      strcpy( item->buffer, item->version );
      strcat( item->buffer, " 100 Continue\r\n\r\n" );

      item->expect = 0;
      item->next = CONTINUE;
      item->count = strlen( item->buffer );
      item->written = 0;
      item->state = HEADER;

      return;
   }

   /*
    * We push all the environment variable data onto the working stack
    * so we can install them in the child process, avoiding the memory
    * leak described in setenv(3) in the parent.  This drove me crazy for
    * two days.  It is necessary to keep previous versions of environment
    * variable values because so that any previously returned pointers to
    * them remain valid for the lifetime of the process.
    */

   if ( item->working == NULL && ( item->working = make_stack()) == NULL )
   {
      remove_conn( item );
      return;
   }

   if ( add_to_env( item, "REMOTE_ADDR" ) ||
        add_to_env( item, item->addr ))
   {
      remove_conn( item );
      return;
   }

   if ( s == NULL && ( s = make_string()) == NULL )
   {
      remove_conn( item );
      return;
   }

   s->free += s->used;
   s->used = 0;
   s->top = s->str;

   for( ptr = item->method; *ptr; ++ptr )
      if ( string_append( s, toupper( *ptr )) == NULL )
         {
            remove_conn( item );
            return;
         }

   if ( add_to_env( item, "REQUEST_METHOD" ) || add_to_env( item, s->str ))
   {
      remove_conn( item );
      return;
   }

   s->free += s->used;
   s->used = 0;
   s->top = s->str;

   if ( string_append( s, '.' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   for( ptr = item->request; *ptr; ++ptr )
      if ( *ptr == '?' )
         break;
      else if ( string_append( s, *ptr ) == NULL )
      {
         remove_conn( item );
         return;
      }

   prog = s->str;
   args = ( *ptr == '?' ? ++ptr : "" );

   if ( stat( prog, &st ) < 0 )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;

      if ( item->type == 2 )
         item->close = 1;

      if ( errno == ENOENT || errno == ENOTDIR )
      {
         item->error = ERR_NOTFOUND;
         return;
      }

      if ( errno == EACCES )
      {
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->error = ERR_INTERNAL;
      return;
   }

   /*
    * Check for file type and permissions.
    */

   if ( ! S_ISREG( st.st_mode ) ||
        (
           ! (( S_IRUSR & st.st_mode ) && passwd->pw_uid == st.st_uid ) &&
           ! (( S_IRGRP & st.st_mode ) && group->gr_gid == st.st_gid ) &&
           ! ( S_IROTH & st.st_mode )
        ) ||
        (
           ! (( S_IXUSR & st.st_mode ) && passwd->pw_uid == st.st_uid ) &&
           ! (( S_IXGRP & st.st_mode ) && group->gr_gid == st.st_gid ) &&
           ! ( S_IXOTH & st.st_mode )
        ))
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;

      if ( item->type == 2 )
         item->close = 1;

      item->error = ERR_FORBIDDEN;
      return;
   }

   /*
    * We refuse to run the CGI program just to generate a header.
    * This response merely indicates the CGI resource exists and is
    * runnable.
    */

   if ( ! item->type )
   {
      item->buffer[ 0 ] = '\0';
      strcpy( item->buffer, item->version );
      strcat( item->buffer, " 200 OK\r\n" );
      add_common_headers( item, 1 );
      strcat( item->buffer, "\r\n" );

      item->written = 0;
      item->count = strlen( item->buffer );
      item->state = HEADER;

      return;
   }

   if ( add_to_env( item, "SCRIPT_NAME" ) ||
        add_to_env( item, prog + 1 ) ||
        add_to_env( item, "QUERY_STRING" ) ||
        add_to_env( item, args ) ||
        add_to_env( item, "SERVER_SOFTWARE" ) ||
        add_to_env( item, "Drood/" VERSION ) ||
        add_to_env( item, "GATEWAY_INTERFACE" ) ||
        add_to_env( item, "CGI/1.1" ) ||
        add_to_env( item, "HTTP_USER_AGENT" ) ||
        add_to_env( item, ( item->agent == NULL ? "" : item->agent )) ||
        add_to_env( item, "HTTP_COOKIE" ) ||
        add_to_env( item, ( item->cookie == NULL ? "" : item->cookie )) ||
        add_to_env( item, "SERVER_NAME" ) ||
        add_to_env( item, item->host ) ||
        add_to_env( item, "CONTENT_TYPE" ) ||
        add_to_env( item, ( item->content == NULL ? "" : item->content )))
   {
      remove_conn( item );
      return;
   }

   if ( item->type == 1 )
   {
      if ( add_to_env( item, "CONTENT_LENGTH" ) || add_to_env( item, "0" ))
      {
         remove_conn( item );
         return;
      }
   }
   else
   {
      if ( item->length == NULL )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_LENGTHREQ;
         return;
      }

      if ( add_to_env( item, "CONTENT_LENGTH" ) ||
           add_to_env( item, item->length ))
      {
         remove_conn( item );
         return;
      }
   }

   /*
    * Will be re-used in filter_cgi_header().
    */

   if ( item->length != NULL )
   {
      free( item->length );
      item->length = NULL;
   }

   *item->buffer = '\0';
   item->count = 0;
   item->written = 0;

   if ( scgi && find_scgi_server( item, prog ))
   {
      item->state = SCGI_CONNECT;
      return;
   }
   else if (( item->source = fork_cgi( prog, item )) < 0 )
   {
      item->state = ERROR;
      item->error = ERR_INTERNAL;
   }
   else
   {
      item->type_found = item->date_found = item->location_found =
         item->length_found = 0;

      item->size = 0;
      item->state = FILTER;
      ev_set( item->source, EVFILT_READ, EV_ADD, item );
      ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
   }

   while(( ptr = stack_pop( item->working )) != NULL )
      free( ptr );
}

void prepare_static( struct ccb *item )
{
   static struct string *s = NULL;
   int len;
   struct stat st;
   struct tm *lt;
   char *ptr, **matches, type[ 128 ], size[ 128 ], modified[ 128 ];

   if ( item->type == 2 )
   {
      item->count = 0;
      item->close = 1;
      item->error = ERR_BADREQUEST;
      item->state = ERROR;
      return;
   }

   if ( s == NULL && ( s = make_string()) == NULL )
   {
      remove_conn( item );
      return;
   }

   s->free += s->used;
   s->used = 0;
   s->top = s->str;

   /*
    * We make sure there is at least one separator between the various
    * items of the path we are going to create in s, because the
    * command-line arguments may not have included them, and the resource
    * specified in the request may have not begun with a slash.  Multiple
    * separators are tolerated by the filesystem-related system calls.
    */

   if ( string_append( s, '.' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   if ( string_append( s, '/' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   for ( ptr = item->host; *ptr; ++ptr )
      if ( string_append( s, *ptr ) == NULL )
      {
         remove_conn( item );
         return;
      }

   if ( string_append( s, '/' ) == NULL )
   {
      remove_conn( item );
      return;
   }

   for( ptr = item->request; *ptr; ++ptr )
      if ( string_append( s, *ptr ) == NULL )
      {
         remove_conn( item );
         return;
      }

   if ( item->gzip )
   {
      if ( string_append( s, '.' ) == NULL )
      {
         remove_conn( item );
         return;
      }

      if ( string_append( s, 'g' ) == NULL )
      {
         remove_conn( item );
         return;
      }

      if ( string_append( s, 'z' ) == NULL )
      {
         remove_conn( item );
         return;
      }
   }

NO_GZIP:
   if ( stat( s->str, &st ) < 0 )
   {
      if ( item->gzip )
      {
         s->top -= 3;
         s->free += 3;
         s->used -= 3;
         s->str[ s->used ] = '\0';

         item->gzip = 0;
         goto NO_GZIP;
      }

      if ( errno == ENOENT || errno == ENOTDIR )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_NOTFOUND;
         return;
      }

      if ( errno == EACCES )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;
      return;
   }

   if ( ! S_ISREG( st.st_mode ))
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   /*
    * Don't send any file which is executable.
    */

   if (( S_IXUSR & st.st_mode ) ||( S_IXGRP & st.st_mode ) ||
         S_IXOTH & st.st_mode )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   if ( ! st.st_size )
   {
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_NOCONTENT;
      return;
   }

   if ( item->type && ( item->source = open( s->str, O_RDONLY )) < 0 )
   {
      if ( errno == EACCES )
      {
         item->count = 0;
         item->written = 0;
         item->state = ERROR;
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, "open(): %m" );
      return;
   }

   ptr = NULL;

   if ( item->gzip )
   {
      s->top -= 3;
      s->free += 3;
      s->used -= 3;
      s->str[ s->used ] = '\0';
   }

   if (( matches = match( &suffix_rx, s->str, s->used, 1 )) == NULL )
   {
      if ( matched )
      {
         remove_conn( item );
         return;
      }
   }
   else
   {
      char **tp;

      for( tp = ( char **)type_table->values; **tp; tp += 2 )
         if ( ! strcmp( *tp, *matches ))
         {
            ptr = *( tp + 1 );
            break;
         }

      free( *matches );
   }

   if ( ptr == NULL )
      ptr = default_type;

   lt = gmtime( &st.st_ctime );
   strftime( modified, sizeof( modified ),
             "Last-Modified: %a, %d %b %Y %H:%M:%S %Z", lt );
   len = strlen( modified );

   modified[ --len ] = 'T';
   modified[ --len ] = 'M';
   modified[ --len ] = 'G';

   if ( item->since != NULL && ! strcmp( item->since, &modified[ 15 ] ))
   {
      item->state = ERROR;
      item->error = ERR_NOTMODIFIED;
      item->count = 0;
      item->written = 0;
      return;
   }

   if ( item->unmodified != NULL && strcmp( item->unmodified, &modified[ 15 ] ))
   {
      item->state = ERROR;
      item->error = ERR_MODIFIED;
      item->count = 0;
      item->written = 0;
      return;
   }

   strcat( modified, "\r\n" );

   /*
    * st.st_size is off_t which we cast to long long unsigned, since
    * there is no printf escape for off_t.
    */

   snprintf( type, sizeof( type ), "Content-Type: %s\r\n", ptr );
   snprintf( size, sizeof( size ), "Content-Length: %llu\r\n",
             ( item->type ? ( long long unsigned int )st.st_size :
                            ( long long unsigned int )0 ));

   *item->buffer = '\0';
   strcpy( item->buffer, item->version );
   strcat( item->buffer, " 200 OK\r\n" );

   strcat( item->buffer, type );
   strcat( item->buffer, size );

   if ( item->gzip )
      strcat( item->buffer, "Content-Encoding: gzip\r\n" );

   strcat( item->buffer, modified );
   add_common_headers( item, 1 );
   strcat( item->buffer, "\r\n" );

   item->count = strlen( item->buffer );
   item->written = 0;
   item->size = st.st_size;
   item->next = RESPOND;
   item->state = HEADER;
}

void analyze_request( struct ccb *item )
{
   int g = 1, h = 1, p = 1, len;
   char *ptr;

   if ( logfd > 0 )
   {
      static time_t t = 0, ot = 0;
      static char buffer[ IOBUFSIZE ], datebuf[ 64 ];
      struct tm *lt;

      if (( t = time( NULL )) != ot )
      {
         ot = t;
         lt = gmtime( &t );
         strftime( datebuf, sizeof( datebuf ),
                   "%a, %d %b %Y %H:%M:%S %Z", lt );
      }

      snprintf( buffer, sizeof( buffer ), "%s|%s|%s|%s|http://%s/%s|%s\n",
                datebuf, item->addr,
                ( item->agent == NULL ? "no user-agent header" : item->agent ),
                item->method,
                ( item->host == NULL ? "default.host" : item->host ),
                ( *item->request == '/' ? &item->request[ 1 ] : item->request ),
                ( item->referer == NULL ? "no referrer header" : item->referer ));

      if ( flock( logfd, LOCK_EX ))
         syslog( LOG_ERR, "flock( logfd, LOCK_EX ): %m" );
      else
      {
         if ( write( logfd, buffer, strlen( buffer )) < 0 )
            syslog( LOG_ERR, "write( logfd ): %m" );

         flock( logfd, LOCK_UN );
      }
   }

   if ( ( item->verno = strcasecmp( item->version, "HTTP/1.0" )) &&
        strcasecmp( item->version, "HTTP/1.1" ))
   {
      item->error = ERR_BADVERSION;
      item->close = 1;
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      return;
   }

   if ( ! item->verno )
      item->close = 1;

   if ( item->host == NULL )
   {
      if (( item->host = str_dup( "default", 7 )) == NULL )
      {
         remove_conn( item );
         return;
      }
   }
   else if ( *item->host == '.' ||
             strchr( item->host, '/' ) != NULL ||
             ( clean_dynamic != NULL && ! strcmp( item->host, clean_dynamic )))
   {
      item->count = 0;
      item->written = 0;
      item->close = 1;
      item->error = ERR_BADREQUEST;
      item->state = ERROR;
      return;
   }

   if ( ( g = strcasecmp( item->method, "GET" )) &&
        ( h = strcasecmp( item->method, "HEAD" )) &&
        ( p = strcasecmp( item->method, "POST" )))
   {
      item->error = ERR_BADMETHOD;
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      return;
   }

   if ( !h )
      item->type = 0;
   else if ( !g )
      item->type = 1;
   else if ( !p )
      item->type = 2;

   /*
    * If the request is a POST, and we're not configured to run (S)CGI,
    * then we bail, closing the connection, to avoid having to read 
    * past the request body to get to the next request header, if this is
    * an HTTP/1.1 connection.  Since errors of this sort are going to
    * be rare because they are most likely the result of a misconfigured
    * server, the simplest solution is just to force the client to open
    * another connection.
    */

   if ( item->type == 2 && dynamic == NULL )
   {
      item->error = ERR_NOTFOUND;
      item->count = 0;
      item->written = 0;
      item->close = 1;
      item->state = ERROR;
   }
   
   if ( !( *item->request ))
   {
      item->error = ERR_BADREQUEST;
      item->count = 0;
      item->written = 0;
      item->close = 1;
      item->state = ERROR;
      return;
   }

   len = strlen( item->request );

   if ( match( &parent_rx, item->request, len, 0 ) != NULL )
   {
      item->error = ERR_BADREQUEST;
      item->count = 0;
      item->written = 0;
      item->state = ERROR;
      item->close = 1;
      return;
   }
      
   if ( item->request[ len - 1 ] == '/' )
   {
      len = strlen( item->request ) + strlen( root_resource );

      if (( ptr = memory( len + 1 )) == NULL )
      {
         remove_conn( item );
         return;
      }

      strcpy( ptr, item->request );
      strcat( ptr, root_resource );

      free( item->request );
      item->request = ptr;
   }

   if ( dynamic != NULL && ! strncmp( dynamic, item->request, dynalen ))
      prepare_dynamic( item );
   else
      prepare_static( item );
}

/*
 * respond_header() and respond_error() are triggered by the writeability
 * of a client socket.  Because we know the socket is writable, we know
 * write() will not generate EWOULDBLOCK, but it can return short counts.
 * If write() returns an error, we drop the connection.
 */

void respond_header( struct ccb *item )
{
   off_t written;

   if ( write_timeout && ! item->timer )
   {
      set_timer( item, 0 );
      item->timer = 1;
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
   {
      item->written = 0;

      if ( ! item->type || (( item->state = item->next ) == HEADER ))
         ( item->close ? remove_conn( item ) : clear_conn( item ) );
      else
      {
         if ( item->state == CHUNK || item->state == NOCHUNK )
         {
            ev_set( item->source, EVFILT_READ, EV_ENABLE, item );
            ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
         }
         else
            respond_static( item );
      }
   }
}

void respond_error( struct ccb *item )
{
   off_t written;

   if ( write_timeout && ! item->timer )
   {
      set_timer( item, 0 );
      item->timer = 1;
   }

   if ( ! item->count )
   {
      char buffer[ 512 ], len_buffer[ 64 ];

      *item->buffer = '\0';

      if ( item->version == NULL )
      {
         strcpy( item->buffer, "HTTP/1.0" );
         item->close = 1;
      }
      else
         strcpy( item->buffer, item->version );

      strcat( item->buffer, errstrings[ item->error ] );
      add_common_headers( item, 1 );

      snprintf( buffer, sizeof( buffer ),
                "<html><head><title>Error</title></head><body>%s</body></html>",
                errstrings[ item->error ] );

      snprintf( len_buffer, sizeof( len_buffer ), "Content-Length: %u\r\n",
                ( unsigned int )strlen( buffer ));

      strcat( item->buffer, len_buffer );
      strcat( item->buffer, "\r\n" );
      strcat( item->buffer, buffer );

      item->count = strlen( item->buffer );
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

/*
 * This handler has been inlined into process_clients(), but is
 * called once for every static resource by respond_header() to
 * transfer the first chunk of response body.
 */

void respond_static( struct ccb *item )
{
   off_t written = 0;

   if ( sendfile( item->source, item->sock, item->written,
                  MIN( item->size, IOBUFSIZE ), NULL,
                  &written, 0 ) < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
      {
         remove_conn( item );
         return;
      }
   }

   item->written += written;
   item->size -= written;

   if ( ! item->size )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

/*
 * Both respond_chunked() and respond_unchunked() are triggered by the
 * readability of CGI pipes for clients in the CHUNK and NOCHUNK states,
 * respectively.  We know the CGI pipe is readable, but we do not know
 * if the client socket is writeable, so we check for EWOULDBLOCK for
 * failed client writes, but not for failed pipe reads.
 */

void respond_unchunked( struct ccb *item )
{
   int result;
   off_t written;

   if ( ! item->count )
   {
      switch(( result = read( item->source, item->buffer, IOBUFSIZE ) ))
      {
         case 0:
            ( item->close ? remove_conn( item ) : clear_conn( item ) );
            return;

         case -1:
            remove_conn( item );
            return;

         default:
            item->count = result;
            item->written = 0;
      }
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         remove_conn( item );

      return;
   }

   item->count -= written;
   item->written += written;
}

void respond_chunked( struct ccb *item )
{
   int result = 1;
   off_t written;

   if ( ! item->count )
   {
      if (( result = read( item->source, &item->buffer[ 6 ], IOBUFSIZE - 8 ))
            < 0 )
      {
         remove_conn( item );
         return;
      }

      item->buffer[ result + 6 ] = '\r';
      item->buffer[ result + 7 ] = '\n';

      item->count = result + 8;
      item->written = 0;

      snprintf( item->buffer, 6, "%.4X\r\n", result );
      item->buffer[ 5 ] = '\n';
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         remove_conn( item );

      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count && ! result )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

void filter_cgi_header( struct ccb *item )
{
   char *line, *ptr, **matches, c;
   int r, notready;

   for( ; ; )
   {
      notready = 0;

      if ( ! item->inactive )
      {
         item->count = 0;
         *item->buffer = '\0';
      }

      for( ; ; )
      {
         r = read( item->source, &c, 1 );

         if ( r <= 0 )
         {
            if ( r < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ))
               notready = 1;

            line = NULL;
            break;
         }

         /*
          * We don't want the line terminators.
          */

         if ( c == 13 )
            continue;

         if ( c == 10 )
         {
            line = item->buffer;
            break;
         }

         /*
          * New character overwrites previous string terminator.  New string
          * terminator added.  Notice that item->count does not count the
          * string terminator.
          */

         item->buffer[ item->count++ ] = c;

         if ( item->count >= IOBUFSIZE )
         {
            notready = 2;
            line = NULL;
            break;
         }

         item->buffer[ item->count ] = '\0';
      }

      if ( line == NULL )
      {
         if ( notready == 1 )
         {
            item->inactive = 1;
            return;
         }
         else
         {
            /*
             * 0 means premature EOF from CGI program.
             * 2 means the CGI program sent a too-long response header line.
             * We bail in each case.
             */

            item->count = 0;
            item->close = 1;
            item->written = 0;
            item->error = ERR_INTERNAL;
            item->state = ERROR;

            flip_events( item );
            return;
         }
      }

      /*
       * If we have the request header separator line, we are done here.
       */

      if ( *line == '\0' )
         break;

      ptr = str_dup( line, item->count );

      /*
       * CGI programs must not create headers larger than the ccb->buffer
       * - 1000.  This give us space to supplement the header without
       * the possibility of buffer overrun.
       */

      if (( item->size += strlen( ptr ) > IOBUFSIZE - 1000 ))
      {
         free( ptr );
         remove_conn( item );
         return;
      }

      if ( stack_push( item->working, ptr ))
      {
         remove_conn( item );
         return;
      }

      if ( ! item->type_found )
      {
         if ( match( &type_rx, line, item->count, 0 ) != NULL )
         {
            item->type_found = 1;
            continue;
         }
      }

      if ( ! item->location_found )
      {
         if ( match( &location_rx, line, item->count, 0 ) != NULL )
         {
            item->location_found = 1;
            continue;
         }
      }

      if ( ! item->date_found )
      {
         if ( match( &date_rx, line, item->count, 0 ) != NULL )
         {
            item->date_found = 1;
            continue;
         }
      }

      if ( ! item->length_found )
      {
         if (( matches = match( &length_rx, line, item->count, 1 )) == NULL )
         {
            if ( matched )
            {
               remove_conn( item );
               return;
            }
         }
         else
         {
            item->length_found = 1;

            if ( item->length != NULL )
               free( item->length );

            item->length = matches[ 0 ];
         }
      }
   }

   item->written = 0;
   *item->buffer = '\0';

   if ( item->location_found )
      item->next = HEADER;
   else if ( ! item->verno || item->length != NULL )
      item->next = NOCHUNK;
   else
      item->next = CHUNK;

   item->state = HEADER;

   strcpy( item->buffer, item->version );
   strcat( item->buffer, ( item->location_found ? " 302 Found\r\n" :
                                                  " 200 OK\r\n" ));

   add_common_headers( item, !item->date_found );

   if ( item->location_found )
      strcat( item->buffer, "Content-Length: 0\r\n" );
   else if ( ! item->type_found )
   {
      strcat( item->buffer, "Content-Type: " );
      strcat( item->buffer, default_type );
      strcat( item->buffer, "\r\n" );
   }

   if ( ! item->location_found && item->next == CHUNK )
      strcat( item->buffer, "Transfer-Encoding: chunked\r\n" );

   if ( item->close )
      strcat( item->buffer, "Connection: close\r\n" );

   while(( line = ( char *)stack_pop( item->working )) != NULL )
   {
      strcat( item->buffer, line );
      strcat( item->buffer, "\r\n" );
      free( line );
   }

   strcat( item->buffer, "\r\n" );
   item->count = strlen( item->buffer );

   non_blocking( item->sock );
   flip_events( item );
}

int accept_connection()
{
   struct sockaddr_storage addr;
   char address[ 128 ];
   const char *ptr = NULL;
   socklen_t len;
   int conn;

   len = sizeof( struct sockaddr_storage );
   conn = accept( fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         syslog( LOG_ERR, "accept(): %m" );

      return 1;
   }

   non_blocking( conn );

   if ( addr.ss_family == AF_INET6 )
      ptr = inet_ntop( AF_INET6, &(( struct sockaddr_in6 *)&addr )->sin6_addr,
                       address, sizeof( address ));
   else if ( addr.ss_family == AF_INET )
      ptr = inet_ntop( AF_INET, &(( struct sockaddr_in *)&addr )->sin_addr,
                       address, sizeof( address ));

   if ( ptr == NULL )
      strcpy( address, "unknown" );

   add_conn( conn, address );
   return 0;
}

void remove_events( int n)
{
   int c;

   for( c = n + 1; c < out; ++c )
      if ( outqueue[ c ].ident == closed )
         outqueue[ c ].ident = -1;
}

void drop_connection( int n )
{
   remove_events( n );
   remove_conn( outqueue[ n ].udata );
}

/*
 * If a timer fires on a descriptor and that descriptor becomes readable
 * after the last invocation of kevent(), we could get the both events in
 * the queue, on the next invocation.  If the read event was earlier in the
 * queue, and in processing the read event, it was decided to close the
 * connection, its ccb would be freed with a call to remove_conn().  When
 * the timer event was encountered a second freeing of the ccb would occur.
 * If the events were placed in the queue in the opposite order, then
 * processing the timer event would call remove_conn(), and then the read
 * event handler would be dispatched with a pointer to the freed ccb.
 * Both situations will crash the server.

 * Our solution is to modify the output queue every time a connection is
 * dropped during the processing of a read, write, or timer event.  We make
 * sure any other events for the same descriptor further along in the queue
 * are ignored by setting the ident member of the appropriate struct kevent
 * to -1.

 * kevent() returns -1 when there is no room to create an error event in
 * the output queue.  We treat this case the same as an error event in the
 * output queue, and just move on.  Errors are generated from duplicate
 * delete timer events in the input queue, and the occasional failure to
 * create a read or a write event on a connection unexpectedly dropped by
 * the client or server.  When kevent() does return -1, the output queue
 * may contain events, but we don't know how many, so we can't service
 * them.  If they are errors, we don't care.  If they are IO events, they
 * are level-triggered, so we will get those events again.  If they are
 * timer events, they will be generated again, for timers are periodic.
 * The worst that can happen is that a timeout will take longer than
 * otherwise to occur.  No events are lost.

 * Because we the listening socket can only appear once in the output queue,
 * the rate at which we accept new connections is inversely proportional to
 * the number of active connections, but we do try and accept up to 100 at
 * once, if they are ready to be accepted.
 */

void process_clients()
{
   int n, kq;

   if (( kq = kqueue()) < 0 )
   {
      syslog( LOG_ERR, "kqueue(): %m" );
      exit( 1 );
   }

   close_on_exec( kq );
   close_on_exec( fd );

   non_blocking( fd );
   ev_set( fd, EVFILT_READ, EV_ADD | EV_ENABLE, NULL );

   for( ; ; )
   {
      if ( restart )
         restart_drood();

      out = kevent( kq, inqueue, in, outqueue, QLEN, NULL );

      in = 0;

      if ( out <= 0 )
         continue;

      for( n = 0; n < out; ++n )
      {
         if ( restart )
            restart_drood();

         if ( outqueue[ n ].ident < 0 || outqueue[ n ].flags & EV_ERROR )
            continue;

         if ( outqueue[ n ].filter == EVFILT_TIMER )
            drop_connection( n );
         else if ( outqueue[ n ].ident == fd )
         {
            int i;

            for( i = 0; i < 100; ++i )
               if ( accept_connection() )
                  break;
         }
         else
         {
            struct ccb *item;
            closed = 0;

            item = ( struct ccb *)outqueue[ n ].udata;

            /*
             * I have inlined the handler for RESPOND for speed.
             */

            if ( item->state != RESPOND )
               item->state( item );
            else
            {
               off_t written = 0;

               if ( sendfile( item->source, item->sock, item->written,
                              MIN( item->size, IOBUFSIZE ), NULL,
                              &written, 0 ) < 0 )
               {
                  if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
                  {
                     remove_conn( item );
                     continue;
                  }
               }

               item->written += written;
               item->size -= written;

               if ( ! item->size )
                  ( item->close ? remove_conn( item ) : clear_conn( item ));
            }

            if ( closed )
               remove_events( n );
         }
      }
   }
}

int fork_children()
{
   pid_t pid;

   if (( children = make_stack()) == NULL )
      exit( 1 );

   while( --instances )
      switch(( pid = fork() ))
      {
         case -1:
            syslog( LOG_ERR, "fork(): %m" );
            /* fall through intended */

         case 0:
            return 1;

         default:
            if ( stack_push_pid( children, pid ))
               exit( 1 );
      }

   return 0;
}

void open_request_log()
{
   if ( logfile == NULL )
      return;

   if (( logfd = open( logfile, O_WRONLY | O_CREAT | O_APPEND )) < 0 )
      syslog( LOG_ERR, "open(): %m" );
   else
      close_on_exec( logfd );
}

void change_directory()
{
   if ( chdir( root ) < 0 )
   {
      syslog( LOG_ERR, "chdir(): %m" );
      exit( 1 );
   }
}

int main( int argc, char **argv )
{
   init_rx();

   set_options( argc, argv );
   set_signals();
   set_types();

   openlog( "drood", LOG_PID, LOG_DAEMON );
   logging = 1;

   if ( !testing )
      become_daemon();

   start_listening();
   child = fork_children();

   open_request_log();
   change_directory();

   /*
    * Parent remains as original user and group so it can restart itself
    * with same permissions.  CGI children call change_identity() before
    * they call execv().
    */

   if ( child )
      change_identity();

   process_clients();

   return 0;
}
